# -*- coding: utf-8 -*-
# wasp_general/network/clients/ftp.py
#
# Copyright (C) 2017 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code
# TODO: write tests for the code

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

import ftplib

from wasp_general.uri import WSchemeSpecification
from wasp_general.network.clients.proto import WNetworkClientProto
from wasp_general.network.clients.base import WBasicNetworkClientProto
from wasp_general.network.clients.base import WBasicNetworkClientListDirCapability
from wasp_general.network.clients.base import WBasicNetworkClientChangeDirCapability
from wasp_general.network.clients.base import WBasicNetworkClientMakeDirCapability
from wasp_general.network.clients.base import WBasicNetworkClientCurrentDirCapability
from wasp_general.network.clients.base import WBasicNetworkClientUploadFileCapability
from wasp_general.network.clients.base import WBasicNetworkClientRemoveFileCapability


class WFTPClient(WBasicNetworkClientProto):

	def __init__(self, uri):
		WBasicNetworkClientProto.__init__(self, uri)

		try:
			ftp_args = {'host': uri.hostname()}
			# TODO: FTP class in python3.6 has port argument. But 3.4 doesn't
			'''
			if uri.port() is not None:
				ftp_args['port'] = uri.port()
			'''

			ftp_client = ftplib.FTP(**ftp_args)

			login_args = {}
			if uri.username() is not None:
				login_args['user'] = uri.username()
			if uri.password():
				login_args['passwd'] = uri.password()
			ftp_client.login(**login_args)

			ftp_client.cwd(uri.path() if uri.path() is not None else '/')

			self.__ftp_client = ftp_client

		except (ftplib.error_perm, ftplib.error_proto, ftplib.error_reply, ftplib.error_temp):
			raise WNetworkClientProto.ConnectionError(
				'Unable to connect to %s' % str(uri)
			)
		except OSError:  # no route to host and so on
			raise WNetworkClientProto.ConnectionError(
				'Unable to connect to %s' % str(uri)
			)

	def ftp_client(self):
		return self.__ftp_client

	def _close(self):
		self.__ftp_client.close()

	@classmethod
	def scheme_specification(cls):
		# TODO: FTP class in python3.6 has port argument. But 3.4 doesn't
		# port = WSchemeSpecification.ComponentDescriptor.optional
		return WSchemeSpecification(
			'ftp',
			username=WSchemeSpecification.ComponentDescriptor.optional,
			password=WSchemeSpecification.ComponentDescriptor.optional,
			hostname=WSchemeSpecification.ComponentDescriptor.required,
			path=WSchemeSpecification.ComponentDescriptor.required,
		)

	@classmethod
	def agent_capabilities(cls):
		return WFTPClientListDirCapability, \
			WFTPClientMakeDirCapability, \
			WFTPClientChangeDirCapability, \
			WFTPClientUploadFileCapability, \
			WFTPClientRemoveFileCapability


class WFTPClientChangeDirCapability(WBasicNetworkClientChangeDirCapability):

	def request(self, path, *args, exec_cmd=None, **kwargs):
		try:
			self.network_agent().ftp_client().cwd(path)
			if exec_cmd is not None:
				return self.network_agent().request(exec_cmd, *args, **kwargs)
			return True
		except (ftplib.error_perm, ftplib.error_proto, ftplib.error_reply, ftplib.error_temp):
			if exec_cmd is not None:
				return
			return False


class WFTPClientCurrentDirCapability(WBasicNetworkClientCurrentDirCapability):

	def request(self, *args, **kwargs):
		return self.network_agent().ftp_client().pwd()


class WFTPClientListDirCapability(WBasicNetworkClientListDirCapability):

	def request(self, *args, **kwargs):
		try:
			return tuple(self.network_agent().ftp_client().nlst())
		except (ftplib.error_perm, ftplib.error_proto, ftplib.error_reply, ftplib.error_temp):
			return


class WFTPClientMakeDirCapability(WBasicNetworkClientMakeDirCapability):

	def request(self, directory_name, *args, **kwargs):
		try:
			return len(self.network_agent().ftp_client().mkd(directory_name)) > 0
		except (ftplib.error_perm, ftplib.error_proto, ftplib.error_reply, ftplib.error_temp):
			return False


class WFTPClientUploadFileCapability(WBasicNetworkClientUploadFileCapability):

	def request(self, file_name, file_obj, *args, **kwargs):
		try:
			self.network_agent().ftp_client().storbinary('STOR ' + file_name, file_obj)
			return True
		except (ftplib.error_perm, ftplib.error_proto, ftplib.error_reply, ftplib.error_temp):
			return False


class WFTPClientRemoveFileCapability(WBasicNetworkClientRemoveFileCapability):

	def request(self, file_name, *args, **kwargs):
		try:
			self.network_agent().ftp_client().delete(file_name)
			return True
		except (ftplib.error_perm, ftplib.error_proto, ftplib.error_reply, ftplib.error_temp):
			return False



# -*- coding: utf-8 -*-
# wasp_general/network/beacon/beacon.py
#
# Copyright (C) 2016 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# TODO: Add zeroconf beacon
# TODO: write tests for the code

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

import os
from enum import Enum
from zmq.eventloop.ioloop import IOLoop
from abc import ABCMeta, abstractmethod

from wasp_general.verify import verify_type
from wasp_general.config import WConfig

from wasp_general.network.primitives import WIPV4SocketInfo
from wasp_general.network.transport import WNetworkNativeTransportProto
from wasp_general.network.beacon.transport import WBroadcastBeaconTransport, WMulticastBeaconTransport
from wasp_general.network.beacon.messenger import WBeaconMessengerBase, WBeaconMessenger
from wasp_general.network.service import WIOLoopService, WNativeSocketHandler, WNativeSocketDirectIOHandler


class WBeaconConfig(metaclass=ABCMeta):
	""" Abstract class that represent service discovery beacon configuration
	"""

	@verify_type(config=(WConfig, None), config_section=(str, None))
	def __init__(self, config=None, config_section=None):
		""" Merge configuration to this beacon

		:param config: configuration storage
		:param config_section: configuration section name where options are
		"""
		self.__configuration = WConfig()
		self.__configuration.merge(os.path.join(os.path.dirname(__file__), '..', '..', 'defaults.ini'))
		if config is not None:
			self.__configuration.merge_section(config, 'wasp-general::network::beacon', section_from=config_section)

	def config(self):
		""" Return beacon configuration

		:return: WConfig
		"""
		return self.__configuration


class WNetworkBeaconCallback(metaclass=ABCMeta):
	""" Abstract class that represent network beacon callback. Helps to interact with clients or servers. After
	beacons message received this is the place where all the beacon logic happens.
	"""

	class WDataDescription(Enum):
		""" This enum defines beacons message source
		"""
		request = 0
		""" Beacons message is request from client to server
		"""
		invalid_request = 1
		""" Beacons message is invalid request from client to server
		"""
		response = 2
		""" Beacons message is response from server to client
		"""

	@abstractmethod
	@verify_type(message=bytes, source=WIPV4SocketInfo, description=WDataDescription)
	def __call__(self, message, source, description):
		""" Method where all the magic is done.

		:param message: binary message (request or response)
		:param source: network source address (original address)
		:param description: defines whether message is a request or a response
		:return: None
		"""
		raise NotImplementedError('This method is abstract')


class WBeaconIOHandler(WNativeSocketDirectIOHandler, metaclass=ABCMeta):
	""" Basic beacon io-handler.
	"""

	message_maxsize = 1024
	""" Network messages maximum size
	"""

	@verify_type(config=WConfig, messanger=(WBeaconMessengerBase, None), callback=(WNetworkBeaconCallback, None))
	def __init__(self, config, messenger=None, callback=None):
		""" Create new io-handler for beacon

		:param config: beacon io-configuration
		:param messenger: beacon messenger (or None for :class:`.WBeaconMessenger`)
		:param callback: beacon callback (or None if it is not required)
		"""
		WNativeSocketDirectIOHandler.__init__(self)
		self.__config = config
		self.__messenger = messenger if messenger is not None else WBeaconMessenger()
		self.__callback = callback

	def messenger(self):
		""" Return beacon messenger

		:return: WBeaconMessenger
		"""
		return self.__messenger

	def max_size(self):
		""" Return maximum message size. (Minimum between this class 'message_maxsize' value and messengers
		'message_maxsize' value)

		:return: int
		"""
		return min(self.message_maxsize, self.messenger().message_maxsize)

	def config(self):
		""" Return handler`s configuration

		:return: WConfig
		"""
		return self.__config

	def callback(self):
		""" Return handler`s callback

		:return: WNetworkBeaconCallback (or None)
		"""
		return self.__callback


class WBeaconHandler(WNativeSocketHandler):
	""" Beacon's loop-handler. Is capable to create required transport (that is specified in the beacon's
	configuration) Depends on configuration value, the following classes is used:

		'broadcast' (default) - :class:`.WBroadcastBeaconTransport`
		'multicast' - :class:`.WMulticastBeaconTransport`
		'unicast_udp' - not implemented yet
		'unicast_tcp' - not implemented yet
	"""

	@verify_type('paranoid', io_handler=WBeaconIOHandler, server_mode=bool)
	@verify_type(config=WConfig, transport=(WNetworkNativeTransportProto, None))
	def __init__(self, config, io_handler, server_mode, transport=None):

		if transport is None:
			transport_cfg = config['wasp-general::network::beacon']['transport'].lower()

			if transport_cfg == 'broadcast':
				transport = WBroadcastBeaconTransport()
			elif transport_cfg == 'multicast':
				transport = WMulticastBeaconTransport()
			elif transport_cfg == 'unicast_udp':
				raise NotImplementedError("This transport doesn't implemented yet")
			elif transport_cfg == 'unicast_tcp':
				raise NotImplementedError("This transport doesn't implemented yet")
			else:
				raise ValueError('Invalid beacon transport type: "%s"' % transport_cfg)

		WNativeSocketHandler.__init__(self, transport, config, io_handler, server_mode)

	@verify_type('paranoid', io_loop=IOLoop)
	def setup_handler(self, io_loop):
		""" :meth:`.WIOLoopServiceHandler.setup_handler` implementation. When this object is in
		'non-server mode' (client mode), then beacon message is sent
		"""
		WNativeSocketHandler.setup_handler(self, io_loop)
		if self.server_mode() is False:
			self.io_handler().transport_socket().sendto(
				self.io_handler().messenger().request(self.config()),
				self.transport().target_socket(self.config()).pair()
			)


class WNetworkBeaconBase(WIOLoopService):
	""" Represent service discovery beacon that works over the network. This beacon doesn't interact with network
	services like zeroconf, but instead it does all the network discovery work itself. The real work is done
	in :class:`.WNetworkServerBeacon` and :class:`.WNetworkClientBeacon` classes.

	This service automatically creates :class:`.WBeaconHandler` object with the specified io-handler.
	"""

	@verify_type('paranoid', config=WConfig, io_handler=WBeaconIOHandler, server_mode=bool)
	@verify_type('paranoid', transport=(WNetworkNativeTransportProto, None), timeout=(int, None))
	def __init__(self, config, io_handler, server_mode, transport, timeout=None):
		""" Create new beacon service

		:param config: beacon's configuration
		:param io_handler: io-handler to use
		:param server_mode: 'server_mode' flag
		:param transport: beacon's transport
		:param timeout: same as timeout in :meth:`.WIOLoopService.__init__`
		"""
		handler = WBeaconHandler(config, io_handler, server_mode, transport)
		WIOLoopService.__init__(self, handler, timeout=timeout)


class WNetworkServerBeacon(WBeaconConfig, WNetworkBeaconBase):
	""" Server beacon that is waiting to respond on a valid request and/or process this request with
	the specified callback.
	"""

	class Handler(WBeaconIOHandler):
		""" Server's handler. Responds on a valid requests, if callback was specified, then it
		process client request. If server has received invalid request and 'process_any' flag was set, then
		callback (if available) will process this request as invalid
		"""

		@verify_type('paranoid', config=WConfig, messanger=(WBeaconMessengerBase, None))
		@verify_type('paranoid', callback=(WNetworkBeaconCallback, None))
		@verify_type(process_any=bool)
		def __init__(self, config, messenger=None, callback=None, process_any=False):
			""" Create new handler

			:param config: same as config in :meth:`.WBeaconIOHandler.__init__`
			:param messenger: same as messenger in :meth:`.WBeaconIOHandler.__init__`
			:param callback: same as callback in :meth:`.WBeaconIOHandler.__init__`
			:param process_any: should callback process invalid requests or not
			"""
			WBeaconIOHandler.__init__(self, config, messenger, callback)
			self.__process_any = process_any

		def process_any(self):
			""" Return 'process_any' flag, that is currently used

			:return: bool
			"""
			return self.__process_any

		def handler_fn(self, fd, event):
			""" :meth:`.WNativeSocketIOHandler.handler_fn` method implementation.
			"""
			s = self.transport_socket()
			messenger = self.messenger()
			callback = self.callback()

			request, client = s.recvfrom(self.max_size())
			original_address = WIPV4SocketInfo(client[0], client[1])

			if messenger.has_response(self.config(), request, original_address) is True:
				direction = WNetworkBeaconCallback.WDataDescription.request

				response = messenger.response(self.config(), request, original_address)
				address = messenger.response_address(self.config(), request, original_address)
				if callback is not None:
					callback(request, original_address, direction)
				s.sendto(response, address.pair())
			elif self.process_any() is True and callback is not None:
				direction = WNetworkBeaconCallback.WDataDescription.invalid_request

				if messenger.valid_response(self.config(), request, original_address):
					callback(request, original_address, direction)

	@verify_type('paranoid', config=(WConfig, None), config_section=(str, None))
	@verify_type('paranoid', messanger=(WBeaconMessengerBase, None), callback=(WNetworkBeaconCallback, None))
	@verify_type('paranoid', process_any=bool, transport=(WNetworkNativeTransportProto, None))
	def __init__(
		self, config=None, config_section=None, messenger=None, callback=None, process_any=False, transport=None
	):
		""" Create new server beacon

		:param config: same as config in :meth:`.WBeaconConfig.__init__`
		:param config_section: same as config_section in :meth:`.WBeaconConfig.__init__`
		:param messenger: same as messenger in :meth:`.WNetworkServerBeacon.Handler.__init__`
		:param callback: same as callback in :meth:`.WNetworkServerBeacon.Handler.__init__`
		:param process_any: same as process_any in :meth:`.WNetworkServerBeacon.Handler.__init__`
		:param transport: same as transport in :meth:`.WNetworkBeaconBase.__init__`
		"""
		WBeaconConfig.__init__(self, config=config, config_section=config_section)
		io_handler = WNetworkServerBeacon.Handler(self.config(), messenger, callback, process_any=process_any)
		WNetworkBeaconBase.__init__(self, self.config(), io_handler, True, transport)


class WNetworkClientBeacon(WBeaconConfig, WNetworkBeaconBase):
	""" Client beacon sends single request and waits for responses for a period of time. This period is specified
	in beacons configuration as 'lookup_timeout' option. Client waiting period can be interrupted by calling
	the :meth:`.WNetworkClientBeacon.stop` method.
	"""

	class Handler(WBeaconIOHandler):
		""" Client handler waits for responses and calls callback if it is available
		"""

		def handler_fn(self, fd, event):
			""" :meth:`.WNativeSocketIOHandler.handler_fn` method implementation.
			"""
			s = self.transport_socket()
			messenger = self.messenger()
			callback = self.callback()
			direction = WNetworkBeaconCallback.WDataDescription.response

			response, server = s.recvfrom(self.max_size())

			if callback is not None:
				server_si = WIPV4SocketInfo(server[0], server[1])
				if messenger.valid_response(self.config(), response, server_si):
					callback(response, server_si, direction)

	@verify_type('paranoid', config=(WConfig, None), config_section=(str, None))
	@verify_type('paranoid', messenger=(WBeaconMessengerBase, None), callback=(WNetworkBeaconCallback, None))
	@verify_type('paranoid', transport=(WNetworkNativeTransportProto, None))
	def __init__(self, config=None, config_section=None, messenger=None, callback=None, transport=None):
		""" Create new client beacon

		:param config: same as config in :meth:`.WBeaconConfig.__init__`
		:param config_section: same as config_section in :meth:`.WBeaconConfig.__init__`
		:param messenger: same as messenger in :meth:`.WNetworkClientBeacon.Handler.__init__`
		:param callback: same as callback in :meth:`.WNetworkClientBeacon.Handler.__init__`
		:param transport: same as transport in :meth:`.WNetworkBeaconBase.__init__`
		"""
		WBeaconConfig.__init__(self, config=config, config_section=config_section)
		io_handler = WNetworkClientBeacon.Handler(self.config(), messenger, callback)
		timeout = self.config().getint('wasp-general::network::beacon', 'lookup_timeout')
		WNetworkBeaconBase.__init__(self, self.config(), io_handler, False, transport, timeout)



# -*- coding: utf-8 -*-

import pytest

from wasp_general.network.messenger.proto import WMessengerOnionProto, WMessengerEnvelopeProto
from wasp_general.network.messenger.proto import WMessengerOnionSessionProto, WMessengerOnionLayerProto
from wasp_general.network.messenger.proto import WMessengerOnionSessionFlowProto


def test_abstract():

	pytest.raises(TypeError, WMessengerOnionProto)
	pytest.raises(NotImplementedError, WMessengerOnionProto.layer, None, '')
	pytest.raises(NotImplementedError, WMessengerOnionProto.layers_names, None)

	pytest.raises(TypeError, WMessengerEnvelopeProto)
	pytest.raises(NotImplementedError, WMessengerEnvelopeProto.message, None)

	pytest.raises(TypeError, WMessengerOnionSessionProto)
	pytest.raises(NotImplementedError, WMessengerOnionSessionProto.onion, None)
	pytest.raises(NotImplementedError, WMessengerOnionSessionProto.process, None, None)

	envelope = TestEnvelopeProto.Envelope()
	session = TestWMessengerOnionSessionProto.Session()
	pytest.raises(TypeError, WMessengerOnionLayerProto)
	pytest.raises(NotImplementedError, WMessengerOnionLayerProto.process, None, envelope, session)

	pytest.raises(TypeError, WMessengerOnionSessionFlowProto)
	pytest.raises(NotImplementedError, WMessengerOnionSessionFlowProto.iterator, None, envelope)


class TestEnvelopeProto:

	class Envelope(WMessengerEnvelopeProto):
		def message(self):
			return

	def test(self):
		e = TestEnvelopeProto.Envelope()
		assert(isinstance(e, TestEnvelopeProto.Envelope) is True)
		assert(isinstance(e, WMessengerEnvelopeProto) is True)
		assert(e.meta() == {})


class TestWMessengerOnionProto:

	class Onion(WMessengerOnionProto):

		def __init__(self):
			self.layers_storage = {}
			for l in [
				TestWMessengerOnionLayerProto.Layer('first_layer'),
				TestWMessengerOnionLayerProto.Layer('l2'),
				TestWMessengerOnionLayerProto.Layer('last')
			]:
				self.layers_storage[l] = l

		def layer(self, layer_name):
			return self.layers_storage[layer_name]

		def layers_names(self):
			return list(self.layers_storage.values())


class TestWMessengerOnionSessionProto:

	class Session(WMessengerOnionSessionProto):

		def onion(self):
			return TestWMessengerOnionProto.Onion()

		def process(self, envelope):
			return


class TestWMessengerOnionLayerProto:

	class Layer(WMessengerOnionLayerProto):

		def process(self, envelope, session, **kwargs):
			e = TestEnvelopeProto.Envelope()
			e.message = lambda: '::' + self.name() + '::' + envelope.message() + '::'
			return e

	def test(self):
		pytest.raises(TypeError, WMessengerOnionLayerProto)

		l = TestWMessengerOnionLayerProto.Layer('layer_name')
		assert(isinstance(l, WMessengerOnionLayerProto) is True)
		assert(l.name() == 'layer_name')
		l = TestWMessengerOnionLayerProto.Layer('l2')
		assert(l.name() == 'l2')

		envelope = TestEnvelopeProto.Envelope()
		envelope.message = lambda: 'msg'
		session = TestWMessengerOnionSessionProto.Session()
		r = l.process(envelope, session)
		assert(isinstance(r, WMessengerEnvelopeProto) is True)
		assert(r.message() == '::l2::msg::')


class TestWMessengerOnionSessionFlow:

	def test(self):
		pytest.raises(TypeError, WMessengerOnionSessionFlowProto.IteratorInfo, 'ln', 4.)

		ii = WMessengerOnionSessionFlowProto.IteratorInfo('layer_name', a=1, b='code')
		assert(ii.layer_name() == 'layer_name')
		assert(ii.layer_args() == {'a': 1, 'b': 'code'})

		pytest.raises(
			TypeError, WMessengerOnionSessionFlowProto.Iterator,
			'ln', 7
		)

		envelope = TestEnvelopeProto.Envelope()

		i1 = WMessengerOnionSessionFlowProto.Iterator('layer')
		assert(isinstance(i1, WMessengerOnionSessionFlowProto.Iterator) is True)
		assert(isinstance(i1, WMessengerOnionSessionFlowProto.IteratorInfo) is True)
		assert(i1.layer_name() == 'layer')
		assert(i1.next(envelope) is None)

		i2 = WMessengerOnionSessionFlowProto.Iterator('layer2', next_iterator=i1)
		assert(i2.layer_name() == 'layer2')
		assert(i2.next(envelope) == i1)



# -*- coding: utf-8 -*-

import pytest

from wasp_general.crypto.aes import WAES, WAESMode, WPKCS7Padding
from wasp_general.crypto.rsa import WRSA

from wasp_general.network.messenger.envelope import WMessengerTextEnvelope, WMessengerBytesEnvelope
from wasp_general.network.messenger.layers import WMessengerOnionCoderLayerProto
from wasp_general.network.messenger.onion import WMessengerOnion
from wasp_general.network.messenger.proto import WMessengerOnionSessionFlowProto
from wasp_general.network.messenger.session import WMessengerOnionSessionFlow, WMessengerOnionSession

from wasp_general.network.messenger.coders import WMessengerFixedModificationLayer, WMessengerEncodingLayer
from wasp_general.network.messenger.coders import WMessengerHexLayer, WMessengerBase64Layer, WMessengerAESLayer
from wasp_general.network.messenger.coders import WMessengerRSALayer


class TestWMessengerFixedModificationLayer:

	def test_layer(self):
		assert(isinstance(WMessengerFixedModificationLayer(), WMessengerFixedModificationLayer) is True)
		assert(isinstance(WMessengerFixedModificationLayer(), WMessengerOnionCoderLayerProto) is True)

		head = WMessengerFixedModificationLayer.Target.head
		tail = WMessengerFixedModificationLayer.Target.tail
		text_envelope = WMessengerTextEnvelope('foo')
		bytes_envelope = WMessengerBytesEnvelope(b'bar')

		onion = WMessengerOnion()
		sf = WMessengerOnionSessionFlow.sequence_flow(WMessengerOnionSessionFlowProto.IteratorInfo('layer'))
		session = WMessengerOnionSession(onion, sf)

		layer = WMessengerFixedModificationLayer()
		result = layer.encode(text_envelope, session, target=head, modification_code='header::')
		assert(isinstance(result, WMessengerTextEnvelope) is True)
		assert(result.message() == 'header::foo')
		pytest.raises(RuntimeError, layer.encode, text_envelope, session, modification_code=b'foo')
		pytest.raises(RuntimeError, layer.decode, bytes_envelope, session, target=tail)
		pytest.raises(TypeError, layer.encode, text_envelope, session, target=head, modification_code=b'foo')
		pytest.raises(TypeError, layer.decode, bytes_envelope, session, target=tail, modification_code='f')
		pytest.raises(TypeError, layer.decode, bytes_envelope, session, target=1, modification_code='f')

		result = layer.decode(result, session, target=head, modification_code='header::')
		assert(isinstance(result, WMessengerTextEnvelope) is True)
		assert(result.message() == 'foo')

		invalid_envelope = WMessengerTextEnvelope('head::foo')
		pytest.raises(
			ValueError, layer.decode, invalid_envelope, session, target=head, modification_code='header::'
		)

		pytest.raises(
			ValueError, layer.decode, invalid_envelope, session, target=head,
			modification_code='very_long_header::'
		)

		result = layer.decode(
			layer.encode(bytes_envelope, session, target=tail, modification_code=b'::tail'),
			session, target=tail, modification_code=b'::tail'
		)
		assert(isinstance(result, WMessengerBytesEnvelope) is True)
		assert(result.message() == b'bar')

		invalid_envelope = WMessengerBytesEnvelope(b'foo::wrong_tail')
		pytest.raises(
			ValueError, layer.decode, invalid_envelope, session, target=tail, modification_code=b'::tail'
		)


class TestWMessengerEncodingLayer:

	def test(self):
		onion = WMessengerOnion()
		sf = WMessengerOnionSessionFlow.sequence_flow(WMessengerOnionSessionFlowProto.IteratorInfo('layer'))
		session = WMessengerOnionSession(onion, sf)
		layer = WMessengerEncodingLayer()

		assert(isinstance(layer, WMessengerEncodingLayer) is True)
		assert(isinstance(layer, WMessengerOnionCoderLayerProto) is True)

		encoded_data = layer.encode(WMessengerTextEnvelope('some text'), session)
		assert(isinstance(encoded_data, WMessengerBytesEnvelope) is True)
		assert(encoded_data.message() == b'some text')

		encoded_data = layer.encode(WMessengerTextEnvelope('some text'), session, encoding='ascii')
		assert(encoded_data.message() == b'some text')

		decoded_data = layer.decode(encoded_data, session, encoding='utf-8')
		assert(isinstance(decoded_data, WMessengerTextEnvelope) is True)
		assert(decoded_data.message() == 'some text')

		decoded_data = layer.decode(encoded_data, session)
		assert(decoded_data.message() == 'some text')


class TestWMessengerHexLayer:

	def test_layer(self):

		onion = WMessengerOnion()
		sf = WMessengerOnionSessionFlow.sequence_flow(WMessengerOnionSessionFlowProto.IteratorInfo('layer'))
		session = WMessengerOnionSession(onion, sf)
		layer = WMessengerHexLayer()

		assert(isinstance(layer, WMessengerHexLayer) is True)
		assert(isinstance(layer, WMessengerOnionCoderLayerProto) is True)

		result = layer.encode(WMessengerBytesEnvelope(b'msg1'), session)
		assert(isinstance(result, WMessengerTextEnvelope) is True)
		assert(result.message() == '6d736731')

		result = layer.decode(WMessengerTextEnvelope('7365636f6e64206d7367'), session)
		assert(isinstance(result, WMessengerBytesEnvelope) is True)
		assert(result.message() == b'second msg')


class TestWMessengerBase64Layer:

	def test_layer(self):

		onion = WMessengerOnion()
		sf = WMessengerOnionSessionFlow.sequence_flow(WMessengerOnionSessionFlowProto.IteratorInfo('layer'))
		session = WMessengerOnionSession(onion, sf)
		layer = WMessengerBase64Layer()

		assert(isinstance(layer, WMessengerBase64Layer) is True)
		assert(isinstance(layer, WMessengerOnionCoderLayerProto) is True)

		result = layer.encode(WMessengerBytesEnvelope(b'msg1'), session)
		assert(isinstance(result, WMessengerBytesEnvelope) is True)
		assert(result.message() == b'bXNnMQ==')

		result = layer.decode(WMessengerBytesEnvelope(b'c2Vjb25kIG1zZw=='), session)
		assert(isinstance(result, WMessengerBytesEnvelope) is True)
		assert(result.message() == b'second msg')


class TestWMessengerAESLayer:

	def test_layer(self):

		onion = WMessengerOnion()
		sf = WMessengerOnionSessionFlow.sequence_flow(WMessengerOnionSessionFlowProto.IteratorInfo('layer'))
		session = WMessengerOnionSession(onion, sf)
		layer = WMessengerAESLayer()

		secret = b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f'
		iv = b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f'

		aes_mode = WAESMode(16, 'AES-CBC', secret + iv, padding=WPKCS7Padding())
		aes_cipher = WAES(aes_mode)

		assert(isinstance(layer, WMessengerAESLayer) is True)
		assert(isinstance(layer, WMessengerOnionCoderLayerProto) is True)

		result = layer.encode(WMessengerBytesEnvelope(b'msg1'), session, aes_cipher=aes_cipher)
		assert(isinstance(result, WMessengerBytesEnvelope) is True)
		assert(result.message() == b'\x14\\O\x1b9e\x853\xfbxx\xfd\xb2\xb4A\xc2')

		result = layer.decode(
			WMessengerBytesEnvelope(b'\x87\xee\xd3^B\x0c\xd1\xc8\xf3va\\`\xbaT\xc9'),
			session, aes_cipher=aes_cipher
		)
		assert(isinstance(result, WMessengerBytesEnvelope) is True)
		assert(result.message() == b'second msg')


class TestWMessengerRSALayer:

	def test_layer(self):

		onion = WMessengerOnion()
		sf = WMessengerOnionSessionFlow.sequence_flow(WMessengerOnionSessionFlowProto.IteratorInfo('layer'))
		session = WMessengerOnionSession(onion, sf)
		layer = WMessengerRSALayer()

		rsa_pk = WRSA.generate_private(1024)

		assert(isinstance(layer, WMessengerRSALayer) is True)
		assert(isinstance(layer, WMessengerOnionCoderLayerProto) is True)

		result = layer.encode(WMessengerBytesEnvelope(b'msg1'), session, public_key=rsa_pk)
		assert(isinstance(result, WMessengerBytesEnvelope) is True)
		assert(len(result.message()) == (1024 / 8))

		result = layer.decode(result, session, private_key=rsa_pk)
		assert(isinstance(result, WMessengerBytesEnvelope) is True)
		assert(result.message() == b'msg1')



# -*- coding: utf-8 -*-

import pytest

from wasp_general.task.registry import WRegisteredTask, WTaskRegistryBase, WTaskRegistryStorage, WTaskRegistry
from wasp_general.task.base import WTask


def test_abstract_classes():
	pytest.raises(TypeError, WTaskRegistryBase)
	pytest.raises(NotImplementedError, WTaskRegistryBase.add, None, None)
	pytest.raises(NotImplementedError, WTaskRegistryBase.remove, None, None)
	pytest.raises(NotImplementedError, WTaskRegistryBase.clear, None)
	pytest.raises(NotImplementedError, WTaskRegistryBase.tasks, None, None)
	pytest.raises(NotImplementedError, WTaskRegistryBase.tasks_by_tag, None, None)
	pytest.raises(NotImplementedError, WTaskRegistryBase.tags, None)
	pytest.raises(NotImplementedError, WTaskRegistryBase.count, None)


class TestWRegisteredTask:

	def test_task(self):

		with pytest.raises(ValueError):
			class T1(metaclass=WRegisteredTask):
				pass

		class FakeR:
			pass

		with pytest.raises(TypeError):
			class T2(metaclass=WRegisteredTask):
				__registry__ = FakeR

		class R(WTaskRegistry):
			__registry_storage__ = WTaskRegistryStorage()

		assert(R.__registry_storage__.count() == 0)

		with pytest.raises(TypeError):
			class T3(metaclass=WRegisteredTask):
				__registry__ = R

		class T4(WTask, metaclass=WRegisteredTask):
			__registry__ = R

		assert(R.__registry_storage__.count() == 0)

		R.__skip_none_registry_tag__ = False

		class T4(WTask, metaclass=WRegisteredTask):
			__registry__ = R

		assert(R.__registry_storage__.count() == 1)

		class T5(WTask, metaclass=WRegisteredTask):
			__registry__ = R
			__auto_registry__ = False
		assert(R.__registry_storage__.count() == 1)


class TestWTaskRegistryStorage:

	def test(self):

		registry_storage = WTaskRegistryStorage()

		class R(WTaskRegistry):
			__registry_storage__ = WTaskRegistryStorage()

		with pytest.raises(TypeError):
			class FakeT1:
				pass
			registry_storage.add(FakeT1)

		with pytest.raises(TypeError):
			class FakeT2(WTask):
				pass
			registry_storage.add(FakeT2)

		class T1(WTask, metaclass=WRegisteredTask):
			__registry__ = R

		class T2(WTask, metaclass=WRegisteredTask):
			__registry__ = R

		assert(registry_storage.count() == 0)
		registry_storage.add(T1)
		assert(registry_storage.count() == 1)
		registry_storage.add(T1)
		assert(registry_storage.count() == 2)
		registry_storage.add(T2)
		assert(registry_storage.count() == 3)

		assert(len(list(filter(lambda x: x == T1, registry_storage.tasks_by_tag(None)))) == 2)
		assert(len(list(filter(lambda x: x == T2, registry_storage.tasks_by_tag(None)))) == 1)

		assert(len(registry_storage.tasks()) == 3)
		assert(len(registry_storage.tasks(task_cls=T2)) == 1)

		registry_storage.clear()
		assert(registry_storage.count() == 0)

		registry_storage.__multiple_tasks_per_tag__ = False
		registry_storage.add(T1)
		assert(registry_storage.count() == 1)

		with pytest.raises(RuntimeError):
			registry_storage.add(T1)

		with pytest.raises(RuntimeError):
			registry_storage.add(T2)

		class T3(WTask, metaclass=WRegisteredTask):
			__registry__ = R
			__registry_tag__ = 1

		registry_storage.add(T3)
		assert(registry_storage.count() == 2)

		assert(registry_storage.tags() == (1, None))

		assert(registry_storage.count() == 2)
		assert(registry_storage.tasks_by_tag(1) == T3)
		assert(registry_storage.tasks_by_tag(None) == T1)
		registry_storage.remove(T3)
		assert(registry_storage.count() == 1)
		assert(registry_storage.tasks_by_tag(None) == T1)
		assert(registry_storage.tasks_by_tag(1) is None)

		assert(registry_storage.tasks_by_tag(2) is None)


class TestWTaskRegistry:

	def test_registry(self):

		class R1(WTaskRegistry):
			__skip_none_registry_tag__ = False

		with pytest.raises(ValueError):
			R1.registry_storage()

		with pytest.raises(TypeError):
			class FakeRS:
				pass
			R1.__registry_storage__ = FakeRS()
			R1.registry_storage()

		R1.__registry_storage__ = WTaskRegistryStorage()
		assert(R1.registry_storage() == R1.__registry_storage__)

		class R2(WTaskRegistry):
			__registry_storage__ = WTaskRegistryStorage()

		class T(WTask, metaclass=WRegisteredTask):
			__registry__ = R2

		assert(R1.registry_storage().count() == 0)
		R1.add(T)
		assert(R1.registry_storage().count() == 1)
		R1.remove(T)
		assert (R1.registry_storage().count() == 0)

		R1.add(T)
		R1.add(T)
		assert (R1.registry_storage().count() == 2)
		R1.clear()
		assert (R1.registry_storage().count() == 0)



# -*- coding: utf-8 -*-
# wasp_general/network/beacon/transport.py
#
# Copyright (C) 2016 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from wasp_general.network.transport import WNetworkNativeTransportSocketConfig, WBroadcastNetworkTransport
from wasp_general.network.transport import WMulticastNetworkTransport


server_configuration = WNetworkNativeTransportSocketConfig('wasp-general::network::beacon', 'bind_address', 'port')
"""
Server side use 'bind_address' options to set up listening socket. Server will be able to receive requests
only at this address. Default is '' (empty string) that works as '0.0.0.0' address, in this case server will
be able to receive requests at any address defined in the system.
"""

client_configuration = WNetworkNativeTransportSocketConfig('wasp-general::network::beacon', 'address', 'port')
"""
Mainly. 'address' and 'port' options are used on client side for destination definition (for server
destination). This options can be used for address validation on server side also. For example,
in :class:`.WMulticastBeaconTransport` class 'address' checks if it is a valid multicast address,
in :class:`.WBroadcastBeaconTransport` class 'address' checks if it is a IPv4 address.
"""


class WBroadcastBeaconTransport(WBroadcastNetworkTransport):
	""" Network beacon transport, that uses IPv4 broadcast communication
	"""

	def __init__(self):
		""" Create new broadcast transport
		"""
		WBroadcastNetworkTransport.__init__(self, client_configuration, server_configuration)


class WMulticastBeaconTransport(WMulticastNetworkTransport):
	""" Network beacon transport, that uses IPv4 multicast communication
	"""

	def __init__(self):
		""" Create new multicast transport
		"""
		WMulticastNetworkTransport.__init__(self, client_configuration, server_configuration)



# -*- coding: utf-8 -*-
# wasp_general/crypto/aes.py
#
# Copyright (C) 2016 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

import io
import re
from Crypto.Cipher import AES as pyAES
from Crypto.Util import Counter
from abc import ABCMeta, abstractmethod

from wasp_general.verify import verify_type, verify_value

from wasp_general.crypto.random import random_int


class WBlockPadding(metaclass=ABCMeta):
	""" Padding/reverse padding class prototype
	"""

	@abstractmethod
	@verify_type(data=bytes, block_size=int)
	@verify_value(block_size=lambda x: x > 0)
	def pad(self, data, block_size):
		""" Pad given data to given size

		:param data: data to pad
		:param block_size: size to pad
		:return: bytes
		"""
		raise NotImplementedError("This method is abstract")

	@abstractmethod
	@verify_type(data=bytes, block_size=int)
	@verify_value(block_size=lambda x: x > 0)
	def reverse_pad(self, data, block_size):
		""" Remove pads and return original data

		:param data: data to remove pads from
		:param block_size: size data aligned to
		:return: bytes
		"""
		raise NotImplementedError("This method is abstract")


class WSimplePadding(WBlockPadding):
	""" Class that pads given data with specified ASCII character
	"""

	@verify_type(padding=(int, None))
	@verify_value(padding=lambda x: x is None or (0 <= x <= 127))
	def __init__(self, padding=None):
		""" Create new padding class

		:param padding: integer code of ASCII character
		"""
		if padding is None:
			padding = 0
		self.__padding_symbol = bytes([padding])

	def padding_symbol(self):
		""" Return character with witch data is padded

		:return: bytes
		"""
		return self.__padding_symbol

	@verify_type(data=bytes, block_size=int)
	@verify_value(block_size=lambda x: x > 0)
	def pad(self, data, block_size):
		""" :meth:`.WBlockPadding.pad` method implementation
		"""
		padding_symbol = self.padding_symbol()

		blocks_count = (len(data) // block_size)
		if (len(data) % block_size) != 0:
			blocks_count += 1

		total_length = blocks_count * block_size
		return self._fill(data, total_length, padding_symbol)

	@verify_type(data=bytes, block_size=int)
	@verify_value(block_size=lambda x: x > 0)
	def reverse_pad(self, data, block_size):
		""" :meth:`.WBlockPadding.reverse_pad` method implementation
		"""
		return data.rstrip(self.padding_symbol())

	@verify_type(data=bytes, total_length=int, padding_symbol=bytes)
	@verify_value(total_length=lambda x: x > 0, padding_symbol=lambda x: len(x) == 1)
	def _fill(self, data, total_length, padding_symbol):
		""" Append padding symbol to the end of data till specified length is reached

		:param data: data to append to
		:param total_length: target length
		:param padding_symbol: symbol to pad
		:return: bytes
		"""
		return data.ljust(total_length, padding_symbol)


class WZeroPadding(WSimplePadding):
	""" Zero padding implementation (just alias for WSimplePadding() object)

	see also: https://en.wikipedia.org/wiki/Padding_(cryptography)#Zero_padding
	"""

	def __init__(self):
		""" Create new padding object
		"""
		WSimplePadding.__init__(self)


class WShiftPadding(WSimplePadding):
	""" Same as :class:`.WSimplePadding` class, but also randomly shifts original data.
	"""

	@verify_type(data=bytes, total_length=int, padding_symbol=bytes)
	@verify_value(total_length=lambda x: x > 0, padding_symbol=lambda x: len(x) == 1)
	def _fill(self, data, total_length, padding_symbol):
		""" Overridden :meth:`.WSimplePadding._fill` method. This methods adds padding symbol at the beginning
		and at the end of the specified data.

		:param data: data to append to
		:param total_length: target length
		:param padding_symbol: symbol to pad
		:return: bytes
		"""

		delta = total_length - len(data)
		return ((padding_symbol * random_int(delta)) + data).ljust(total_length, padding_symbol)

	@verify_type(data=bytes, block_size=int)
	@verify_value(block_size=lambda x: x > 0)
	def reverse_pad(self, data, block_size):
		""" :meth:`.WBlockPadding.reverse_pad` method implementation
		"""
		padding_symbol = self.padding_symbol()
		return data.lstrip(padding_symbol).rstrip(padding_symbol)


class WPKCS7Padding(WBlockPadding):
	""" PKCS7 Padding implementation

	see also: https://en.wikipedia.org/wiki/Padding_(cryptography)#PKCS7
	"""

	@verify_type(data=bytes, block_size=int)
	@verify_value(block_size=lambda x: x > 0)
	def pad(self, data, block_size):
		""" :meth:`.WBlockPadding.pad` method implementation
		"""
		pad_byte = block_size - (len(data) % block_size)
		return data + bytes([pad_byte] * pad_byte)

	@verify_type(data=bytes, block_size=int)
	@verify_value(data=lambda x: len(x) > 0, block_size=lambda x: x > 0)
	def reverse_pad(self, data, block_size):
		""" :meth:`.WBlockPadding.reverse_pad` method implementation
		"""
		pad_byte = data[-1]
		if pad_byte > block_size:
			raise ValueError('Invalid padding')

		padding = bytes([pad_byte] * pad_byte)
		if data[-pad_byte:] != padding:
			raise ValueError('Invalid padding')

		return data[:-pad_byte]


class WAESMode:
	""" This class specifies modes of AES encryption. It describes secret key (size and value), block cipher mode
	of operation, padding object (:class:`.WBlockPadding` instance), required initialization values. Note,
	padding is required if source data isn't aligned to block size.

	For byte-sequence generation (that is used as secret key and initialization values) it is possible to use
	:class:`wasp_general.crypto.kdf.WPBKDF2`. :class:`wasp_general.crypto.kdf.WPBKDF2` is a wrapper for PBKDF2
	function (KDF function that safely generates byte-sequence from the given password and salt)

	Currently, only two cipher mode of operation are implemented: 'CBC' and 'CTR'

	see also: https://en.wikipedia.org/wiki/Block_cipher_mode_of_operation
	"""

	__data_padding_length__ = pyAES.block_size
	""" Length to which data must be aligned (in bytes)
	"""
	__init_vector_length__ = pyAES.block_size
	""" Initialization vector length (in bytes)
	"""
	__counter_size__ = pyAES.block_size
	""" Initialization counter size (in bytes)
	"""

	__mode_re__ = re.compile('AES(\-|_)(\\d+)(\-|_)(\\w+)')
	""" Regular expression for parsing cipher name
	"""

	__valid_key_sizes__ = (16, 24, 32)
	""" Supported AES key sizes (in bytes)
	"""

	__modes_descriptor__ = {
		'AES-CBC': {
			'mode_code': pyAES.MODE_CBC,
			'requirements': {
				'initialization_vector': True,
				'counter': False
			}
		},
		'AES-CTR': {
			'mode_code': pyAES.MODE_CTR,
			'requirements': {
				'initialization_vector': False,
				'counter': True
			}
		}
	}
	""" Describes block cipher modes of operation and theirs requirements
	"""

	class SequenceChopper:
		""" Helper, that chops the given byte-sequence into several separate objects (like secret key,
		initialization vector or initialization counter values). The exact values depend on AES key size and
		block cipher mode of operation.

		If length of the given byte-sequence is greater then it is required, then extra bytes discard and
		this extra-bytes don't take part in any calculation
		"""

		@verify_type('paranoid', block_cipher_mode=str, key_size=int)
		@verify_type(sequence=bytes)
		@verify_value('paranoid', block_cipher_mode=lambda x: x in WAESMode.__modes_descriptor__.keys())
		def __init__(self, key_size, block_cipher_mode, sequence):
			""" Create new chopper

			:param key_size: AES secret length
			:param block_cipher_mode: name of block cipher mode of operation
			:param sequence: byte-sequence to chop
			"""
			required_length = self.required_sequence_length(key_size, block_cipher_mode)
			self.__key_size = key_size
			self.__mode = block_cipher_mode
			self.__sequence = sequence

			if required_length > 0:
				if len(self.__sequence) < required_length:
					raise ValueError(
						'Initialization byte-sequence too short. '
						'Must be at least %i bytes long' % required_length
					)

		def secret(self):
			""" Return AES secret generated from the initial byte-sequence

			:return: bytes
			"""
			return self.__sequence[:self.__key_size]

		def initialization_vector(self):
			""" Return initialization vector generated from the initial byte-sequence if it is required
			by the current block cipher mode of operation. If it doesn't require - then None is returned

			:return: bytes or None
			"""
			req = self.__requirements()
			if req['initialization_vector'] is not True:
				return None
			start_position = self.__key_size
			end_position = start_position + WAESMode.__init_vector_length__
			return self.__sequence[start_position:end_position]

		def initialization_counter_value(self):
			""" Return initialization counter value generated from the initial byte-sequence if it is
			required by the current block cipher mode of operation. If it doesn't require - then None
			is returned

			:return: int or None
			"""
			req = self.__requirements()
			if req['counter'] is not True:
				return None

			start_position = self.__key_size
			if req['initialization_vector'] is True:
				start_position += WAESMode.__init_vector_length__
			end_position = start_position + WAESMode.__counter_size__
			seq = self.__sequence[start_position:end_position]
			return int.from_bytes(seq, byteorder='big')

		def __requirements(self):
			""" Return requirements specification (just shortcut to access specific mode requirements from
			WAESMode.__modes_descriptor__)

			:return: dict
			"""
			return WAESMode.__modes_descriptor__[self.__mode]['requirements']

		@classmethod
		@verify_type(key_size=int, block_cipher_mode=str)
		@verify_value(key_size=lambda x: x in WAESMode.__valid_key_sizes__)
		@verify_value(block_cipher_mode=lambda x: x in WAESMode.__modes_descriptor__.keys())
		def required_sequence_length(cls, key_size, block_cipher_mode):
			""" Calculate required byte-sequence length

			:param key_size: AES secret length
			:param block_cipher_mode: name of block cipher mode of operation to calculate for

			:return: int
			"""
			req = WAESMode.__modes_descriptor__[block_cipher_mode]['requirements']
			result = key_size
			if req['initialization_vector'] is True:
				result += WAESMode.__init_vector_length__
			if req['counter'] is True:
				result += WAESMode.__counter_size__
			return result

	@verify_type(key_size=int, block_cipher_mode=str, padding=(None, WBlockPadding), init_sequence=bytes)
	@verify_value(key_size=lambda x: x in WAESMode.__valid_key_sizes__)
	@verify_value(block_cipher_mode=lambda x: x in WAESMode.__modes_descriptor__.keys())
	def __init__(
		self, key_size, block_cipher_mode, init_sequence, padding=None
	):
		""" Create new AES-mode.

		:param key_size: secret length
		:param block_cipher_mode: name of block cipher mode of operation
		:param padding: padding object (if required)
		:param init_sequence: AES secret with initialization vector or counter value
		"""
		self.__key_size = key_size
		self.__mode = block_cipher_mode
		self.__padding = padding
		self.__sequence_chopper = WAESMode.SequenceChopper(key_size, block_cipher_mode, init_sequence)
		self.__cipher_args = (self.__sequence_chopper.secret(),)
		self.__cipher_kwargs = {}

		cipher_descriptor = WAESMode.__modes_descriptor__[block_cipher_mode]
		self.__cipher_kwargs['mode'] = cipher_descriptor['mode_code']

		iv = self.__sequence_chopper.initialization_vector()
		if iv is not None:
			self.__cipher_kwargs['IV'] = iv
		counter = self.__sequence_chopper.initialization_counter_value()
		if counter is not None:
			self.__cipher_kwargs['counter'] = Counter.new(
				WAESMode.__counter_size__ * 8, initial_value=counter
			)

	def key_size(self):
		""" Return cipher secret key size

		:return: int
		"""
		return self.__key_size

	def mode(self):
		""" Return block cipher mode of operation name

		:return:
		"""
		return self.__mode

	def padding(self):
		""" Return padding object

		:return: WBlockPadding or None
		"""
		return self.__padding

	def initialization_vector(self):
		""" Return currently used initialization vector or None if vector is not used

		:return: bytes or None
		"""
		return self.__sequence_chopper.initialization_vector()

	def initialization_counter_value(self):
		""" Return currently used initialization counter value or None if counter is not used

		:return: int or None
		"""
		return self.__sequence_chopper.initialization_counter_value()

	def pyaes_args(self):
		""" Generate and return position-dependent arguments, that are used in :meth:`.AES.new` method

		:return: tuple
		"""
		return self.__cipher_args

	def pyaes_kwargs(self):
		""" Generate and return position-independent (named) arguments, that are used in :meth:`.AES.new` method

		:return: dict
		"""
		return self.__cipher_kwargs

	@classmethod
	def init_sequence_length(cls, key_size, block_cipher_mode):
		""" Return required byte-sequence length

		:param key_size: secret size
		:param block_cipher_mode: name of block cipher mode of operation

		:return: int
		"""
		return WAESMode.SequenceChopper.required_sequence_length(key_size, block_cipher_mode)

	@classmethod
	@verify_type(name=str)
	def parse_cipher_name(cls, name):
		""" Parse cipher name (name like 'aes_256_cbc' or 'AES-128-CTR'). Also this method validates If the
		cipher is supported by this class. If no - exception is raised

		:param name: name to parse

		:return: tuple where the first element is a key size in bytes (int) and the second element - block cipher mode
		of operation (str) (for example: (16, 'AES-CTR') or (24, 'AES-CBC'))
		"""
		r = cls.__mode_re__.match(name.upper())
		if r is None:
			raise ValueError('Unable to find suitable cipher for: "%s"' % name)
		key_size = int(int(r.group(2)) / 8)
		block_mode = 'AES-%s' % r.group(4)
		if key_size not in cls.__valid_key_sizes__:
			raise ValueError('Unsupported secret length: "%i"' % key_size)
		if block_mode not in cls.__modes_descriptor__.keys():
			raise ValueError('Unsupported block cipher mode of operation: "%s"' % block_mode)
		return key_size, block_mode


class WAES:
	""" PyCrypto AES-encryption wrapper
	"""

	@verify_type(mode=WAESMode)
	def __init__(self, mode):
		""" Create new AES cipher with specified mode

		:param mode: AES mode
		"""

		self.__mode = mode

	def mode(self):
		""" Return AES mode

		:return: WAESMode
		"""
		return self.__mode

	def cipher(self):
		""" Generate AES-cipher

		:return: Crypto.Cipher.AES.AESCipher
		"""
		cipher = pyAES.new(*self.mode().pyaes_args(), **self.mode().pyaes_kwargs())
		return cipher

	@verify_type(data=(str, bytes))
	def encrypt(self, data):
		""" Encrypt the given data with cipher that is got from AES.cipher call.

		:param data: data to encrypt
		:return: bytes
		"""
		padding = self.mode().padding()
		if padding is not None:
			data = padding.pad(data, WAESMode.__data_padding_length__)

		return self.cipher().encrypt(data)

	@verify_type(data=bytes, decode=bool)
	def decrypt(self, data, decode=False):
		""" Decrypt the given data with cipher that is got from AES.cipher call.

		:param data: data to decrypt
		:param decode: whether to decode bytes to str or not
		:return: bytes or str (depends on decode flag)
		"""

		result = self.cipher().decrypt(data)

		padding = self.mode().padding()
		if padding is not None:
			result = padding.reverse_pad(result, WAESMode.__data_padding_length__)

		return result.decode() if decode else result



# -*- coding: utf-8 -*-

import os
from wasp_general.mime import mime_type


def test_mime():
	path = os.path.dirname(__file__)
	assert(mime_type(os.path.join(path, 'mime_test_files', 'test.html')) == 'text/html')
	assert(
		mime_type(os.path.join(path, 'mime_test_files', 'test.js')) in
		['application/x-javascript', 'application/javascript']
	)
	assert(mime_type(os.path.join(path, 'mime_test_files', 'test.css')) == 'text/css')
	assert(mime_type(os.path.join(path, 'mime_test_files', 'test.woff2')) == 'application/font-woff2')



# -*- coding: utf-8 -*-

import pytest

from mako.template import Template
from mako.lookup import TemplateCollection

from wasp_general.template import WTemplate, WMakoTemplate, WTemplateText, WTemplateFile
from wasp_general.template import WTemplateLookup, WTemplateRenderer


def test_abstract():
	pytest.raises(TypeError, WTemplate)
	pytest.raises(NotImplementedError, WTemplate.template, None)


class TestWMakoTemplate:

	def test_template(self):
		t = Template(text='code')
		wr = WMakoTemplate(t)
		assert(isinstance(wr, WMakoTemplate) is True)
		assert(isinstance(wr, WTemplate) is True)
		assert(wr.template() == t)


class TestWTemplateText:

	def test_template(self):
		t = WTemplateText('template code')
		assert(isinstance(t, WTemplateText) is True)
		assert(isinstance(t, WTemplate) is True)
		assert(t.template().render() == 'template code')


class TestWTemplateFile:

	def test_template(self, tmpdir):
		f = tmpdir.join('tmp')
		f.write('template')

		t = WTemplateFile(f.strpath)
		assert(isinstance(t, WTemplateFile) is True)
		assert(isinstance(t, WTemplate) is True)
		assert(t.template().render() == 'template')


class TestWTemplateLookup:

	class Collection(TemplateCollection):

		def get_template(self, uri, relativeto=None):
			return WTemplateText('tmpl: ' + uri).template()

	def test_template(self):
		collection = TestWTemplateLookup.Collection()
		t = WTemplateLookup('uri1', collection)

		assert(isinstance(t, WTemplateLookup) is True)
		assert(isinstance(t, WTemplate) is True)
		assert(t.template().render() == 'tmpl: uri1')


class TestWTemplateRenderer:

	def test_renderer(self):
		template = WTemplateText('test code')
		renderer = WTemplateRenderer(template)
		assert(renderer.template() == template.template())
		assert(renderer.context() == {})
		assert(renderer.render() == 'test code')

		template = WTemplateText('var a is ${a}')
		renderer = WTemplateRenderer(template, context={'a': 1})
		assert(renderer.render() == 'var a is 1')

		renderer.update_context(a='2')
		assert(renderer.render() == 'var a is 2')



# -*- coding: utf-8 -*-
# wasp_general/cli/cli.py
#
# Copyright (C) 2016 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code
# TODO: write tests for the code

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

import traceback

from abc import ABCMeta, abstractmethod
from copy import deepcopy

from wasp_general.verify import verify_type, verify_value
from wasp_general.command.command import WCommandSet, WCommandResultProto


class WConsoleHistory:
	""" Simple console history implementation
	"""
	def __init__(self):
		self.__history = []
		self.__history_position = None

	def size(self):
		""" Returns history entries count

		:return: int
		"""
		return len(self.__history)

	@verify_type(pos=(int, None))
	@verify_value(pos=lambda x: x is None or x >= 0)
	def position(self, pos=None):
		""" Get current and/or set history cursor position

		:param pos: if value is not None, then current position is set to pos and new value is returned

		:return: int or None (if position have never been changed)
		"""
		if pos is not None:
			if pos >= len(self.__history):
				raise IndexError('History position is out of bound')
			self.__history_position = pos
		return self.__history_position

	@verify_type(value=str)
	def add(self, value):
		""" Add new record to history. Record will be added to the end

		:param value: new record
		:return: int record position in history
		"""
		index = len(self.__history)
		self.__history.append(value)
		return index

	@verify_type(position=int)
	def entry(self, position):
		""" Get record from history by record position

		:param position: record position
		:return: str
		"""
		return self.__history[position]

	@verify_type(value=str, position=(int, None))
	@verify_value(position=lambda x: x is None or x >= 0)
	def update(self, value, position):
		""" Change record in this history

		:param value: new record to save
		:param position: record position to change
		:return: None
		"""
		self.__history[position] = value


class WConsoleProto(metaclass=ABCMeta):
	""" Basic class for console implementation. It has non-changeable and changeable history
	(:class:`.WConsoleHistory`). One stores previous entered rows, other one helps to entered new row by editing
	previous one.
	"""

	def __init__(self):
		self.__history_mode = False
		self.__history = WConsoleHistory()
		self.__editable_history = None
		self.__current_row = None
		self.__prompt_show = None

	def history(self):
		""" Return changeable history

		:return: WConsoleHistory or None
		"""
		return self.__editable_history

	@verify_type(mode_value=(bool, None))
	def history_mode(self, mode_value=None):
		""" Get and/or set current history mode.

		History mode defines what row will be changed with :meth:`.WConsoleProto.update_row` or can be got by
		:meth:`.WConsoleProto.row` call. If history mode disabled, then :meth:`.WConsoleProto.update_row` and
		:meth:`.WConsoleProto.row` affects current row prompt. If history mode is enabled, then
		:meth:`.WConsoleProto.update_row` and :meth:`.WConsoleProto.row` affects current entry in
		history :class:`.WConsoleHistory` (entry at :meth:`.WConsoleHistory.position`)

		History mode is turned off by default.

		:param mode_value: True value enables history mode. False - disables. None - do nothing
		:return: bool
		"""
		if mode_value is not None:
			self.__history_mode = mode_value
		return self.__history_mode

	def start_session(self):
		""" Start new session and prepare environment for new row editing process

		:return: None
		"""
		self.__current_row = ''
		self.__history_mode = False
		self.__editable_history = deepcopy(self.__history)
		self.__prompt_show = True
		self.refresh_window()

	def fin_session(self):
		""" Finalize current session

		:return: None
		"""
		self.__prompt_show = False
		self.__history.add(self.row())
		self.exec(self.row())

	@verify_type(value=str)
	def update_row(self, value):
		""" Change row

		:param value: new row
		:return: None
		"""
		if not self.__history_mode:
			self.__current_row = value
		else:
			self.history().update(value, self.history().position())

	def row(self):
		""" Get row

		:return: str
		"""
		if not self.__history_mode:
			return self.__current_row
		else:
			return self.history().entry(self.history().position())

	def prompt_show(self):
		""" Return flag, that shows, whether to display prompt and current command at the window end, or not

		:return: bool
		"""
		return self.__prompt_show

	@abstractmethod
	def prompt(self):
		""" Return prompt, that would be printed before row. Prompt length must be the same
		within every session

		:return: str
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	def refresh_window(self):
		""" Refresh current screen. Simple clear and redraw should work

		:return: None
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	@verify_type(row=str)
	def exec(self, row):
		""" Must execute given command

		:param row: command to execute
		:return: None
		"""
		raise NotImplementedError('This method is abstract')


class WConsoleWindowProto(metaclass=ABCMeta):
	""" Basic class for console window implementation
	"""

	@verify_type(console=WConsoleProto)
	def __init__(self, console):
		""" Create new console window

		:param console: console, that this window is linked to
		"""
		self.__console = console
		self.__previous_data = ''
		self.__cursor_position = 0

		if self.width() < 2:
			raise RuntimeError('Invalid width. Minimum windows width is 2')

		if self.height() < 2:
			raise RuntimeError('Invalid height. Minimum windows height is 2')

	@abstractmethod
	def width(self):
		""" Get window width. If windows width was changed - window must be refreshed via
		:meth:`.WConsoleWindowProto.refresh`

		:return: int
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	def height(self):
		""" Get window height. If windows height was changed - window must be refreshed via
		:meth:`.WConsoleWindowProto.refresh`

		:return: int
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	def clear(self):
		""" Clear window and remove every symbol it has

		:return: None
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	@verify_type(line_index=int, line=str)
	@verify_value(line_index=lambda x: x >= 0)
	def write_line(self, line_index, line):
		""" Write string on specified line

		:param line_index: line index to display
		:param line: string to display (must fit windows width)
		:return:
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	@verify_type(y=int, x=int)
	@verify_value(x=lambda x: x >= 0, y=lambda x: x >= 0)
	def set_cursor(self, y, x):
		""" Set input cursor in window to specified coordinates. 0, 0 - is top left coordinates

		:param y: vertical coordinates, 0 - top, bottom - positive value
		:param x: horizontal coordinates, 0 - left, right - positive value
		:return:
		"""
		raise NotImplementedError('This method is abstract')

	@verify_type(prompt_show=bool)
	def refresh(self, prompt_show=True):
		""" Refresh current window. Clear current window and redraw it with one of drawers

		:param prompt_show: flag, that specifies, whether to show prompt and current row at the
		windows end, or not

		:return: None
		"""
		raise NotImplementedError('This method is abstract')

	@verify_type(previous_data=bool, prompt=bool, console_row=bool, console_row_to_cursor=bool)
	@verify_type(console_row_from_cursor=bool)
	def data(
		self, previous_data=False, prompt=False, console_row=False,
		console_row_to_cursor=False, console_row_from_cursor=False
	):
		""" Return output data. Flags specifies what data to append. If no flags was specified
		nul-length string returned

		:param previous_data: If True, then previous output appends
		:param prompt: If True, then console prompt appends. If console_row or console_row_to_cursor is True, \
		then this value is omitted
		:param console_row: If True, then console prompt and current input appends.
		:param console_row_to_cursor: If True, then console prompt and current input till cursor appends. \
		If console_row is True, then this value is omitted
		:param console_row_from_cursor: If True, then current input from cursor appends. \
		If console_row is True, then this value is omitted
		:return: str
		"""

		result = ''

		if previous_data:
			result += self.__previous_data

		if prompt or console_row or console_row_to_cursor:
			result += self.console().prompt()

		if console_row or (console_row_from_cursor and console_row_to_cursor):
			result += self.console().row()
		elif console_row_to_cursor:
			result += self.console().row()[:self.cursor()]
		elif console_row_from_cursor:
			result += self.console().row()[self.cursor():]

		return result

	@verify_type('paranoid', previous_data=bool, prompt=bool, console_row=bool, console_row_to_cursor=bool)
	@verify_type('paranoid', console_row_from_cursor=bool)
	def list_data(
		self, previous_data=False, prompt=False, console_row=False,
		console_row_to_cursor=False, console_row_from_cursor=False
	):
		""" Return list of strings. Where each string is fitted to windows width. Parameters are the same as
		they are in :meth:`.WConsoleWindow.data` method

		:return: list of str
		"""
		return self.split(self.data(
			previous_data, prompt, console_row, console_row_to_cursor, console_row_from_cursor
		))

	def console(self):
		""" Return linked console

		:return: WConsoleProto
		"""
		return self.__console

	@verify_type('paranoid', start_position=int)
	@verify_value('paranoid', start_position=lambda x: x >= 0)
	@verify_type(data=list)
	def write_data(self, data, start_position=0):
		""" Write data from the specified line

		:param data: string to write, each one on new line
		:param start_position: starting line
		:return:
		"""
		if len(data) > self.height():
			raise ValueError('Data too long (too many strings)')

		for i in range(len(data)):
			self.write_line(start_position + i, data[i])

	@verify_type(pos=(None, int))
	@verify_value(pos=lambda x: x is None or x >= 0)
	def cursor(self, pos=None):
		""" Set and/or get relative cursor position. Defines cursor position in current input row.

		:param pos: if value is not None, then current cursor position is set to this value and the same value \
		is returned
		:return: int
		"""
		if pos is not None:
			self.__cursor_position = pos

		return self.__cursor_position

	def commit(self):
		""" Store current input row. Keep current input row as previous output

		:return: None
		"""
		self.__previous_data += (self.data(console_row=True) + '\n')

	@verify_type(data=str)
	def split(self, data):
		""" Split data into list of string, each (self.width() - 1) length or less. If nul-length string
		specified then empty list is returned

		:param data: data to split
		:return: list of str
		"""
		line = deepcopy(data)
		line_width = (self.width() - 1)

		lines = []
		while len(line):

			new_line = line[:line_width]

			new_line_pos = new_line.find('\n')
			if new_line_pos >= 0:
				new_line = line[:new_line_pos]
				line = line[(new_line_pos + 1):]
			else:
				line = line[line_width:]

			lines.append(new_line)

		return lines

	@verify_type(feedback=str, cr=bool)
	def write_feedback(self, feedback, cr=True):
		""" Store feedback. Keep specified feedback as previous output

		:param feedback: data to store
		:param cr: whether to write carriage return to the end or not
		:return: None
		"""
		self.__previous_data += feedback
		if cr is True:
			self.__previous_data += '\n'

	@verify_type(length=int)
	@verify_value(length=lambda x: x >= 0)
	def truncate_feedback(self, length):
		""" Remove data from feedback (removes text from previous output)

		:param length: string length to remove (including required cr-characters)
		:return: None
		"""
		self.__previous_data = self.__previous_data[:-length]


class WConsoleDrawerProto(metaclass=ABCMeta):
	""" Basic class that helps displaying console content
	"""

	@abstractmethod
	@verify_type(window=WConsoleWindowProto, prompt_show=bool)
	def suitable(self, window, prompt_show=True):
		""" Check if this class can display console content

		:param window: window that should be drawn
		:param prompt_show: same as 'prompt_show' in :meth:`.WConsoleWindowProto.refresh` method
		:return: bool (True if this class can draw console content, False - if it can not)
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	@verify_type(window=WConsoleWindowProto, prompt_show=bool)
	def draw(self, window, prompt_show=True):
		""" Display console content on console window

		:param window: windows to draw
		:param prompt_show: same as 'prompt_show' in :meth:`.WConsoleWindowProto.refresh` method
		:return: None
		"""
		raise NotImplementedError('This method is abstract')


class WConsoleWindowBase(WConsoleWindowProto, metaclass=ABCMeta):
	""" Basic class for console window implementation
	"""

	@verify_type(console=WConsoleProto, drawers=WConsoleDrawerProto)
	def __init__(self, console, *drawers):
		"""
		:param drawers: drawers to use
		"""
		WConsoleWindowProto.__init__(self, console)
		self.__drawers = []
		self.__drawers.extend(drawers)

	@verify_type('paranoid', prompt_show=bool)
	def refresh(self, prompt_show=True):
		""" Refresh current window. Clear current window and redraw it with one of drawers

		:param prompt_show: flag, that specifies, whether to show prompt and current row at the
		windows end, or not

		:return: None
		"""
		self.clear()
		for drawer in self.__drawers:
			if drawer.suitable(self, prompt_show=prompt_show):
				drawer.draw(self, prompt_show=prompt_show)
				return

		raise RuntimeError('No suitable drawer was found')


class WConsoleBase(WConsoleProto):

	@verify_type(command_set=(WCommandSet, None))
	def __init__(self, command_set=None):
		WConsoleProto.__init__(self)
		self.__command_set = command_set if command_set is not None else WCommandSet()

	def command_set(self):
		return self.__command_set

	def window(self):
		raise NotImplementedError('This method is abstract')

	def start_session(self):
		""" :meth:`.WConsoleProto.start_session` implementation. Sets cursor to 0 position before session

		:return: None
		"""
		self.window().cursor(0)
		WConsoleProto.start_session(self)

	def fin_session(self):
		""" :meth:`.WConsoleProto.fin_session` implementation. Commits current input row

		:return: None
		"""
		self.window().commit()
		WConsoleProto.fin_session(self)

	@verify_type('paranoid', result=str, cr=bool)
	def write(self, result, cr=True):
		"""  Shortcut for self.window().write_feedback(result) call

		:param result: same as feedback in :meth:`WConsoleWindowProto.write_feedback`
		:param cr: same as cr in :meth:`WConsoleWindowProto.write_feedback`
		:return: None
		"""
		self.window().write_feedback(result, cr=cr)

	@verify_type('paranoid', length=int)
	@verify_value('paranoid', length=lambda x: x >= 0)
	def truncate(self, length):
		"""  Shortcut for self.window().truncate_feedback(result) call

		:param length: same as length in :meth:`WConsoleWindowProto.truncate_feedback`
		:return: None
		"""
		self.window().truncate_feedback(length)

	def refresh_window(self):
		""" Shortcut for self.window().refresh() call

		:return: None
		"""
		self.window().refresh(prompt_show=self.prompt_show())

	def prompt(self):
		""" :meth:`.WConsoleProto.prompt` implementation
		"""
		return '> '

	@verify_type(result=WCommandResultProto)
	def handle_result(self, result):
		result = str(result)
		if len(result) > 0:
			self.write(result)

	@verify_type(e=Exception)
	def handle_exception(self, e):
		if isinstance(e, WCommandSet.NoCommandFound):
			self.write('Error: no suitable command found')
		else:
			self.write('Internal error. Traceback and exception information:')
			self.write(traceback.format_exc())

	@verify_type('paranoid', row=str)
	def exec(self, row):
		command_set = self.command_set()
		try:
			self.handle_result(command_set.exec(row))
		except Exception as e:
			self.handle_exception(e)




import os
from setuptools import setup, find_packages

from wasp_general.version import __package_data__


def read(fname):
	return open(os.path.join(os.path.dirname(__file__), fname)).read()


__pypi_data__ = __package_data__['pypi']


if __name__ == "__main__":
	setup(
		name=__package_data__['package'],
		version=__package_data__['numeric_version'],
		author=__package_data__['author'],
		author_email=__package_data__['author_email'],
		maintainer=__package_data__['maintainer'],
		maintainer_email=__package_data__['maintainer_email'],
		description=__package_data__['brief_description'],
		license=__package_data__['license'],
		keywords=__pypi_data__['keywords'],
		url=__package_data__['homepage'],
		packages=find_packages(),
		include_package_data=\
			__pypi_data__['include_package_data'] if 'include_package_data' in __pypi_data__ else True,
		long_description=read(__package_data__['readme_file']),
		classifiers=__pypi_data__['classifiers'],
		install_requires=read('requirements.txt').splitlines(),
		zip_safe=__pypi_data__['zip_safe'] if 'zip_safe' in __pypi_data__ else False,
		scripts=__package_data__['scripts'] if 'scripts' in __package_data__ else [],
		extras_require=__pypi_data__['extra_require'] if 'extra_require' in __pypi_data__ else {}
	)



# -*- coding: utf-8 -*-
# wasp_general/network/messenger/auth.py
#
# Copyright (C) 2016 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code
# TODO: write tests for the code

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__


from enum import Enum
from abc import ABCMeta, abstractmethod

from wasp_general.verify import verify_type, verify_value

'''
from wasp_general.network.messenger.onion import WMessengerOnionBase, WMessengerOnionLayer, WMessageModifier
from wasp_general.network.messenger.onion import WMessengerOnionSessionBase


class WMessengerAuth(metaclass=ABCMeta):

	class Action(Enum):
		permit = 0
		forbid = 1

	@verify_type(
		message=(bytes, str, None), onion=WMessengerOnionBase,
		session=WMessengerOnionSessionBase, current_layer_index=int
	)
	@verify_value(current_layer_index=lambda x: x >= -1)
	def pack(self, message, onion, session, current_layer_index):
		raise NotImplementedError('This method is abstract')

	@verify_type(
		message=(bytes, str, None), onion=WMessengerOnionBase,
		session=WMessengerOnionSessionBase, current_layer_index=int
	)
	@verify_value(current_layer_index=lambda x: x >= -1)
	def unpack(self, message, onion, session, current_layer_index):
		raise NotImplementedError('This method is abstract')


class WMessengerAuthLayer(WMessengerOnionLayer):

	__layer_name__ = ""

	@verify_type(auth=WMessengerAuth)
	def __init__(self, auth):
		WMessengerOnionLayer.__init__(self, WMessengerAuthLayer.__layer_name__)
		self.__auth = auth

	@verify_type(
		message=(bytes, str, None), onion=WMessengerOnionBase,
		session=WMessengerOnionSessionBase, current_layer_index=int
	)
	@verify_value(current_layer_index=lambda x: x >= -1)
	def pack(self, message, onion, session, current_layer_index):
		pass

	@verify_type(
		message=(bytes, str, None), onion=WMessengerOnionBase,
		session=WMessengerOnionSessionBase, current_layer_index=int
	)
	@verify_value(current_layer_index=lambda x: x >= -1)
	def unpack(self, message, onion, session, current_layer_index):
		pass
'''



# -*- coding: utf-8 -*-
# wasp_general/<FILENAME>.py
#
# Copyright (C) 2018 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code
# TODO: write tests for the code

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from abc import ABCMeta, abstractmethod
from enum import Enum
from threading import Event, Semaphore, Thread
from select import select
from datetime import timedelta

from wasp_general.verify import verify_type
from wasp_general.network.service_v3.proto import WPollingHandlerProto


class WSelectPollingHandler(WPollingHandlerProto):

	def polling_function(self):
		r_list = []
		w_list = []
		x_list = []

		fds = self.file_obj()
		event_mask = self.event_mask()

		if event_mask >> 0 & 1:
			r_list.extend(fds)
		if event_mask >> 1 & 1:
			w_list.extend(fds)
		x_list.extend(fds)

		function_args = (r_list, w_list, x_list, self.timeout())

		def result():
			r_list, w_list, x_list = select(
				function_args[0], function_args[1], function_args[2], function_args[3]
			)
			if len(x_list) > 0:
				raise WPollingHandlerProto.PollingError(*x_list)

			if len(r_list) == 0 and len(w_list) == 0:
				return

			return r_list, w_list

		return result


'''
class WSimpleClientConnectionHandler(WClientConnectionHandlerProto):

	def __init__(self):
		pass




class WSimpleConnectionHandler(WAcceptConnectionHandlerProto):

	def __init__(self, maximum_threads):
		self.__maximum_threads = maximum_threads
		self.__semaphore = Semaphore(maximum_threads)
		self.__threads = []

	def maximum_threads(self):
		return self.__maximum_threads

	def semaphore(self):
		return self.__semaphore

	def handle_connection(self, socket_obj):

		def threading_fn():
			try:
				print('read from socket')
				print(socket_obj)
				client = socket_obj.accept()

			finally:
				self.semaphore().release()

		self.semaphore().acquire()
		th = Thread(target=threading_fn)
		th.start()
		self.__threads.append(th)
'''


class WNetworkService(metaclass=ABCMeta):

	__default_timeout__ = 1

	@verify_type(timeout=(int, None))
	def __init__(self, socket_obj, service_factory, stop_event=None, timeout=None, polling_handler=None):
		self.__socket_obj = socket_obj
		self.__service_factory = service_factory
		self.__timeout = timeout if timeout is not None else self.__default_timeout__
		self.__stop_event = stop_event if stop_event is not None else Event()
		self.__polling_handler = polling_handler if polling_handler is not None else WSelectPollingHandler()

	def socket(self):
		return self.__socket_obj

	def service_factory(self):
		return self.__service_factory

	def timeout(self):
		return self.__timeout

	def stop_event(self):
		return self.__stop_event

	def polling_handler(self):
		return self.__polling_handler

	def start(self):
		service_factory = self.service_factory()

		polling_handler = self.polling_handler()
		polling_handler.setup_poll(WPollingHandlerProto.PollingEvent.read, self.timeout())
		polling_handler.poll_fd(self.socket())

		polling_fn = polling_handler.polling_function()

		while self.stop_event().is_set() is False:
			polling_result = polling_fn()
			if polling_result is not None:
				connection = polling_result[0][0].accept()
				try:
					worker = service_factory.worker(self.stop_event())
					worker.process(connection[0])
				except RuntimeError:
					pass

	def stop(self):
		self.stop_event().set()



# -*- coding: utf-8 -*-

import pytest
from datetime import datetime

from wasp_general.task.thread import WThreadTask
from wasp_general.datetime import utc_datetime

from wasp_general.task.scheduler.proto import WScheduleTask, WScheduleRecord, WTaskSourceProto
from wasp_general.task.scheduler.proto import WRunningRecordRegistryProto, WSchedulerServiceProto


def test_abstract():
	pytest.raises(TypeError, WTaskSourceProto)
	pytest.raises(NotImplementedError, WTaskSourceProto.has_records, None)
	pytest.raises(NotImplementedError, WTaskSourceProto.next_start, None)
	pytest.raises(NotImplementedError, WTaskSourceProto.tasks_planned, None)
	pytest.raises(NotImplementedError, WTaskSourceProto.scheduler_service, None)
	pytest.raises(TypeError, WRunningRecordRegistryProto)

	schedule = WScheduleRecord(TestWScheduleTask.DummyTask())
	pytest.raises(NotImplementedError, WRunningRecordRegistryProto.exec, None, schedule)
	pytest.raises(NotImplementedError, WRunningRecordRegistryProto.running_records, None)
	pytest.raises(TypeError, WSchedulerServiceProto)
	pytest.raises(NotImplementedError, WSchedulerServiceProto.update, None)


class TestWScheduleTask:

	class DummyTask(WScheduleTask):

		def thread_started(self):
			pass

		def thread_stopped(self):
			pass

	def test(self):
		task = TestWScheduleTask.DummyTask()
		assert(isinstance(task, WScheduleTask) is True)
		assert(isinstance(task, WThreadTask) is True)

		assert(task.stop_event() is not None)
		assert(task.ready_event() is not None)

		assert(task.uid() is not None)
		assert(TestWScheduleTask.DummyTask().uid() != TestWScheduleTask.DummyTask().uid())


class TestWScheduleRecord:

	def test(self):
		task = TestWScheduleTask.DummyTask()

		pytest.raises(TypeError, WScheduleRecord, task, policy=1)

		schedule = WScheduleRecord(task)
		assert(schedule.task() == task)
		assert(schedule.policy() == WScheduleRecord.PostponePolicy.wait)
		assert(schedule.task_group_id() is None)
		assert(schedule.task_uid() == task.uid())

		drop_callback_result = []

		def drop_callback():
			drop_callback_result.append(1)

		wait_callback_result = []

		def wait_callback():
			wait_callback_result.append(1)

		schedule = WScheduleRecord(task, on_drop=drop_callback)
		assert(drop_callback_result == [])
		schedule.task_dropped()
		assert(drop_callback_result == [1])

		schedule = WScheduleRecord(task, on_wait=wait_callback)
		assert(wait_callback_result == [])
		schedule.task_postponed()
		assert(wait_callback_result == [1])



# -*- coding: utf-8 -*-
# wasp_general/network/beacon/messenger.py
#
# Copyright (C) 2016 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# TODO: Merge with wasp_general.network.messenger

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

import re
from abc import ABCMeta, abstractmethod

from wasp_general.verify import verify_type, verify_value
from wasp_general.config import WConfig

from wasp_general.network.primitives import WIPV4SocketInfo, WIPPort


class WBeaconMessengerBase(metaclass=ABCMeta):
	""" This is interface for classes, that implement communication (messaging) logic for beacons

	see also: :class:`.WNetworkBeacon`
	"""
	message_maxsize = 512

	@abstractmethod
	@verify_type(beacon_config=WConfig)
	def request(self, beacon_config):
		""" Generate client request for beacon. It is calling from client side.

		:param beacon_config: client beacon configuration.
		:return: bytes
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	@verify_type(beacon_config=WConfig, request=bytes, client_address=WIPV4SocketInfo)
	def has_response(self, beacon_config, request, client_address):
		""" Whether this messenger has response or it must skip the request. Return True if there is a response,
		otherwise - False. If this method returns False, then calling a :meth:`.WBeaconMessengerBase.response`
		method treats as error.

		:param beacon_config:
		:param request:
		:param client_address:
		:return: bool
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	@verify_type(beacon_config=WConfig, request=bytes, client_address=WIPV4SocketInfo)
	def response(self, beacon_config, request, client_address):
		""" Generate server response for clients request. Obviously, it is calling from server side

		:param beacon_config: server beacon configuration
		:param request: client request message
		:param client_address: client address
		:return: bytes
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	@verify_type(beacon_config=WConfig, request=bytes, client_address=WIPV4SocketInfo)
	def response_address(self, beacon_config, request, client_address):
		""" Return client address where server must send response. It is Possible, that address where server
		must send response is different then the origin address. In that case, client address
		is encoded in request message.

		:param beacon_config: server configuration
		:param request: client request message
		:param client_address: original client address
		:return: WIPV4SocketInfo
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	@verify_type(beacon_config=WConfig, response=bytes, server_address=WIPV4SocketInfo)
	def valid_response(self, beacon_config, response, server_address):
		""" Return True if server response isn't junk.

		:param beacon_config: client beacon configuration
		:param response: server response
		:param server_address: original server address
		:return: bool
		"""
		raise NotImplementedError('This method is abstract')


class WBeaconMessenger(WBeaconMessengerBase):
	""" Simple. Demo/debug messenger. Server side just return original hello message to the sender.
	"""

	__beacon_hello_msg__ = b'HELLO'
	""" Client request
	"""

	@verify_type('paranoid', beacon_config=WConfig)
	def request(self, beacon_config):
		""" :meth:`.WBeaconMessengerBase.request` method implementation.
		Sends :attr:`.WBeaconMessenger.__beacon_hello_msg__` to a server
		"""
		return self.__beacon_hello_msg__

	@verify_type('paranoid', beacon_config=WConfig, client_address=WIPV4SocketInfo)
	@verify_type(request=bytes)
	def has_response(self, beacon_config, request, client_address):
		""" :meth:`.WBeaconMessengerBase.has_response` method implementation. This class has a response only if
		the request wasn't empty
		"""
		return True if len(request) > 0 else False

	@verify_type('paranoid', beacon_config=WConfig, client_address=WIPV4SocketInfo)
	@verify_type(request=bytes)
	def response(self, beacon_config, request, client_address):
		""" :meth:`.WBeaconMessengerBase.response` method implementation.
		In response sends the same data as server has got
		"""
		return request

	@verify_type('paranoid', beacon_config=WConfig, request=bytes)
	@verify_type(client_address=WIPV4SocketInfo)
	def response_address(self, beacon_config, request, client_address):
		""" :meth:`.WBeaconMessengerBase.request` method implementation.
		Return the same address, that server detects (assume that clients address is correct)
		"""
		return client_address

	@verify_type('paranoid', beacon_config=WConfig, response=bytes, server_address=WIPV4SocketInfo)
	@verify_type(response=bytes)
	def valid_response(self, beacon_config, response, server_address):
		""" :meth:`.WBeaconMessengerBase.valid_response` method implementation. Response is valid if it has
		anything.
		"""
		return True if len(response) > 0 else False


class WBeaconGouverneurMessenger(WBeaconMessengerBase):
	""" Basic and real messenger implementation. Request and response are generated the same way, but have different
	meaning. Messages are generated the following way:
	[Message header]<:[Address<:TCP/UDP port>]>
	"Message header" is the first part of the message and it must be exactly the same for client or server.
	If server has got message with header that doesn't match its own header, such messages will be omitted.
	Next parts are "Address" and "TCP/UDP port". These parts are separated
	by the :attr:`.WBeaconGouverneurMessenger.__message_splitter__` separator. "TCP/UDP port" can't be specified
	without an "Address" An "Address" field can be an IP address or a domain name.

	For the client "Address" and "UDP Port" are treated as address where server must send the response.
	But for the server, they are treated as the address, that the server publish. For both type of messages
	"Address" and "TCP/UDP port" are generated from a configuration. Options are located
	in 'wasp-general::network::beacon' section. Configuration option "public_address" is used as "Address"
	value and option "public_port" is used as "UDP Port"
	"""

	__message_splitter__ = b':'
	""" Delimiter, that is used for header, IP address and port separation.
	"""

	@verify_type(hello_message=bytes, invert_hello=bool)
	@verify_value(hello_message=lambda x: len(x.decode('ascii')) >= 0)
	@verify_value(hello_message=lambda x: WBeaconGouverneurMessenger.__message_splitter__ not in x)
	def __init__(self, hello_message, invert_hello=False):
		""" Construct new messenger

		:param hello_message: Message header
		:param invert_hello: this flag defines whether response will have the original header \
		('hello_message' value) or it will have reversed value. For example, when this flag is set to True \
		and 'hello_message' is b'sample', then response will have b'elpmas' header.
		"""
		WBeaconMessengerBase.__init__(self)
		self.__gouverneur_message = hello_message
		self.__invert_hello = invert_hello

	def invert_hello(self):
		""" Return whether this messenger was constructed with 'invert_hello' or not.

		:return: bool
		"""
		return self.__invert_hello

	@verify_type(invert_hello=bool)
	def hello_message(self, invert_hello=False):
		""" Return message header.

		:param invert_hello: whether to return the original header (in case of False value) or reversed \
		one (in case of True value).
		:return: bytes
		"""
		if invert_hello is False:
			return self.__gouverneur_message

		hello_message = []
		for i in range(len(self.__gouverneur_message) - 1, -1, -1):
			hello_message.append(self.__gouverneur_message[i])
		return bytes(hello_message)

	@verify_type('paranoid', beacon_config=WConfig, invert_hello=bool)
	def _message(self, beacon_config, invert_hello=False):
		""" Generate request/response message.

		:param beacon_config: server or client configuration. Client configuration is used for request and \
		server configuration for response
		:param invert_hello: return message with reverse header when this argument is set to True.
		:return: bytes
		"""

		message = self.hello_message(invert_hello=invert_hello) + self._message_address_generate(beacon_config)
		return message

	@verify_type(beacon_config=WConfig)
	def _message_address_generate(self, beacon_config):
		""" Generate address for request/response message.

		:param beacon_config: server or client configuration. Client configuration is used for request and \
		server configuration for response
		:return: bytes
		"""

		address = None

		if beacon_config['wasp-general::network::beacon']['public_address'] != '':
			address = str(WIPV4SocketInfo.parse_address(
				beacon_config['wasp-general::network::beacon']['public_address']
			)).encode('ascii')

		if address is not None:
			address = WBeaconGouverneurMessenger.__message_splitter__ + address

			if beacon_config['wasp-general::network::beacon']['public_port'] != '':
				port = beacon_config.getint('wasp-general::network::beacon', 'public_port')
				address += WBeaconGouverneurMessenger.__message_splitter__ + str(port).encode('ascii')

		return address if address is not None else b''

	@verify_type(message=bytes, invert_hello=bool)
	def _message_address_parse(self, message, invert_hello=False):
		""" Read address from beacon message. If no address is specified then "nullable" WIPV4SocketInfo returns

		:param message: message to parse
		:param invert_hello: defines whether message header is the original one or reversed.
		:return: WIPV4SocketInfo
		"""
		message_header = self.hello_message(invert_hello=invert_hello)

		if message[:len(message_header)] != message_header:
			raise ValueError('Invalid message header')

		message = message[len(message_header):]
		message_parts = message.split(WBeaconGouverneurMessenger.__message_splitter__)

		address = None
		port = None

		if len(message_parts) > 3:
			raise ValueError('Invalid message. Too many separators')
		elif len(message_parts) == 3:
			address = WIPV4SocketInfo.parse_address(message_parts[1].decode('ascii'))
			port = WIPPort(int(message_parts[2]))
		elif len(message_parts) == 2 and len(message_parts[1]) > 0:
			address = WIPV4SocketInfo.parse_address(message_parts[1].decode('ascii'))

		return WIPV4SocketInfo(address, port)

	@verify_type('paranoid', beacon_config=WConfig)
	def request(self, beacon_config):
		""" :meth:`.WBeaconMessengerBase.request` method implementation.

		see :class:`.WBeaconGouverneurMessenger`
		"""
		return self._message(beacon_config)

	@verify_type('paranoid', beacon_config=WConfig, request=bytes, client_address=WIPV4SocketInfo)
	def has_response(self, beacon_config, request, client_address):
		""" :meth:`.WBeaconMessengerBase.has_response` method implementation. This method compares request
		header with internal one.
		"""
		try:
			self._message_address_parse(request, invert_hello=self.__invert_hello)
			return True
		except ValueError:
			pass
		return False

	@verify_type('paranoid', beacon_config=WConfig, request=bytes, client_address=WIPV4SocketInfo)
	def response(self, beacon_config, request, client_address):
		""" :meth:`.WBeaconMessengerBase.request` method implementation.

		see :class:`.WBeaconGouverneurMessenger`
		"""
		return self._message(beacon_config, invert_hello=self.__invert_hello)

	@verify_type('paranoid', beacon_config=WConfig, request=bytes)
	@verify_type(client_address=WIPV4SocketInfo)
	def response_address(self, beacon_config, request, client_address):
		""" :meth:`.WBeaconMessengerBase.request` method implementation.

		see :class:`.WBeaconGouverneurMessenger`
		"""
		si = self._message_address_parse(request, invert_hello=self.__invert_hello)
		address = si.address()
		port = si.port()
		return WIPV4SocketInfo(
			address if address is not None else client_address.address(),
			port if port is not None else client_address.port()
		)

	@verify_type('paranoid', beacon_config=WConfig, response=bytes, server_address=WIPV4SocketInfo)
	def valid_response(self, beacon_config, response, server_address):
		""" :meth:`.WBeaconMessengerBase.valid_response` method implementation. Response when it has correct
		header. Response header must be reversed if this messenger was constructed with 'invert_hello' flag.
		"""
		try:
			self._message_address_parse(response, invert_hello=self.__invert_hello)
			return True
		except ValueError:
			pass
		return False


class WHostgroupBeaconMessenger(WBeaconGouverneurMessenger):
	""" This messenger is based on :class:`.WBeaconGouverneurMessenger` class. This messenger extends
	:class:`.WBeaconGouverneurMessenger` functionality by working with host group names. Also, this class doubles
	maximum message size defined in :class:`.WBeaconMessengerBase`. In cases where no host group name are
	specified this class works as its basic class :class:`.WBeaconGouverneurMessenger`.

	Messages format is updated from this:

	'[Message header]<:[Address<:TCP/UDP port>]>'

	to this:

	'[Message header]<:[Address<:TCP/UDP port>]><#<[Hostgroup name]>,<[Hostgroup name]...>'>

	'Message header', 'Address' and 'TCP/UDP port' work the same way as they specified in
	:class:`.WBeaconGouverneurMessenger` class. At the end of the original message new optional separator is
	appended. This separator is defined at :attr:`.WHostgroupBeaconMessenger.__message_groups_splitter__`.
	After this separator host group names can be defined. Host group names can contain latin letters and numbers
	only. Host group names are separated by :attr:`.WHostgroupBeaconMessenger.__group_splitter__`

	If this messenger is constructed with at least one host group name, then it won'not response to requests,
	that do not have at least one host group name, that this messenger was constructed with.
	"""

	message_maxsize = 1024
	""" Doubled message size for list of names
	"""
	__message_groups_splitter__ = b'#'
	""" Bytes that separate original message from extended one
	"""
	__group_splitter__ = b','
	""" Bytes that separate different names
	"""

	re_hostgroup_name = re.compile('^[a-zA-Z0-9]+$')
	""" Regular expression for host group name validation
	"""

	@verify_type('paranoid', hello_message=bytes, invert_hello=bool)
	@verify_value('paranoid', hello_message=lambda x: len(x.decode('ascii')) >= 0)
	@verify_value('paranoid', hello_message=lambda x: WHostgroupBeaconMessenger.__message_groups_splitter__ not in x)
	@verify_type(hostgroup_names=str)
	@verify_value(hostgroup_names=lambda x: WHostgroupBeaconMessenger.re_hostgroup_name.match(x) is not None)
	def __init__(self, hello_message, *hostgroup_names, invert_hello=False):
		""" Create new messenger

		:param hello_message: same as hello_message in :meth:`.WBeaconGouverneurMessenger.__init__`
		:param hostgroup_names: list of host group names
		"""
		WBeaconGouverneurMessenger.__init__(self, hello_message, invert_hello=invert_hello)

		self.__hostgroups = []
		self.__hostgroups.extend([x.encode() for x in hostgroup_names])

	def hostgroups(self):
		""" Return list of host group names

		:return: list of str
		"""
		return [x.decode() for x in self.__hostgroups]

	@verify_type('paranoid', beacon_config=WConfig, invert_hello=bool)
	def _message(self, beacon_config, invert_hello=False):
		""" Overridden :meth:`.WBeaconGouverneurMessenger._message` method. Appends encoded host group names
		to requests and responses.

		:param beacon_config: beacon configuration
		:return: bytes
		"""
		m = WBeaconGouverneurMessenger._message(self, beacon_config, invert_hello=invert_hello)
		hostgroups = self._message_hostgroup_generate()
		if len(hostgroups) > 0:
			m += (WHostgroupBeaconMessenger.__message_groups_splitter__ + hostgroups)
		return m

	def _message_hostgroup_generate(self):
		""" Encode messenger host group names
		:return: bytes
		"""
		return b','.join(self.__hostgroups)

	@verify_type(message=bytes)
	def _message_hostgroup_parse(self, message):
		""" Parse given message and return list of group names and socket information. Socket information
		is parsed in :meth:`.WBeaconGouverneurMessenger._message_address_parse` method

		:param message: bytes
		:return: tuple of list of group names and WIPV4SocketInfo
		"""
		splitter_count = message.count(WHostgroupBeaconMessenger.__message_groups_splitter__)
		if splitter_count == 0:
			return [], WBeaconGouverneurMessenger._message_address_parse(self, message)
		elif splitter_count == 1:
			splitter_pos = message.find(WHostgroupBeaconMessenger.__message_groups_splitter__)
			groups = []
			group_splitter = WHostgroupBeaconMessenger.__group_splitter__
			for group_name in message[(splitter_pos + 1):].split(group_splitter):
				groups.append(group_name.strip())
			address = WBeaconGouverneurMessenger._message_address_parse(self, message[:splitter_pos])
			return groups, address
		else:
			raise ValueError('Invalid message. Too many separators')

	@verify_type('paranoid', beacon_config=WConfig, request=bytes, client_address=WIPV4SocketInfo)
	def has_response(self, beacon_config, request, client_address):
		""" :meth:`.WBeaconMessengerBase.has_response` method implementation. This method compares request
		headers as :meth:`.WBeaconGouverneurMessenger.has_response` do and compares specified group names
		with internal names.
		"""
		try:
			groups, address = self._message_hostgroup_parse(request)

			if len(self.__hostgroups) == 0 or len(groups) == 0:
				return True

			for group_name in groups:
				if group_name in self.__hostgroups:
					return True
			return False
		except ValueError:
			pass
		return False

	@verify_type('paranoid', beacon_config=WConfig, request=bytes)
	@verify_type(client_address=WIPV4SocketInfo)
	def response_address(self, beacon_config, request, client_address):
		""" :meth:`.WBeaconMessengerBase.response_address` method implementation. It just removes host group names
		part and return :meth:`.WBeaconMessengerBase.response_address` result
		"""
		si = self._message_hostgroup_parse(request)[1]
		address = si.address()
		port = si.port()
		return WIPV4SocketInfo(
			address if address is not None else client_address.address(),
			port if port is not None else client_address.port()
		)



# -*- coding: utf-8 -*-

import pytest

from wasp_general.types.binarray import WBinArray

from wasp_general.network.primitives import WMACAddress, WIPV4Address, WNetworkIPV4, WNetworkIPV4Iterator, WFQDN
from wasp_general.network.primitives import WIPPort, WIPV4SocketInfo


class TestWMACAddress:

	def test_from_string(self):

		pytest.raises(ValueError, WMACAddress, "01:23:45:ab")
		pytest.raises(ValueError, WMACAddress, "01:23:45:ab:cd:ZZ")

		assert(
			str(WMACAddress("01:23:45:ab:cd:ef").bin_address()) ==
			"000000010010001101000101101010111100110111101111"
		)
		assert(
			str(WMACAddress.from_string("01-23-45-ab-cd-ef").bin_address()) ==
			"000000010010001101000101101010111100110111101111"
		)
		assert(
			str(WMACAddress.from_string("01:23:45:ab:cd:ef").bin_address()) ==
			"000000010010001101000101101010111100110111101111"
		)
		assert(
			str(WMACAddress.from_string("0123.45ab.cdef").bin_address()) ==
			"000000010010001101000101101010111100110111101111"
		)
		assert(
			str(WMACAddress.from_string("012345abcdef").bin_address()) ==
			"000000010010001101000101101010111100110111101111"
		)

	def test_to_string(self):
		assert(
			str(WMACAddress(WBinArray(0b000000010010001101000101101010111100110111101111))) ==
			"01:23:45:ab:cd:ef"
		)

	def test_to_bytes(self):
		assert(
			bytes(WMACAddress(WBinArray(0b000000010010001101000101101010111100110111101111))) ==
			b"\x01\x23\x45\xab\xcd\xef"
		)


class TestWIPV4Address:

	def test_init(self):
		pytest.raises(TypeError, WIPV4Address, 0.1)
		assert(isinstance(WIPV4Address("10.192.0.1"), WIPV4Address) is True)
		assert(isinstance(WIPV4Address(WBinArray(1)), WIPV4Address) is True)
		assert(isinstance(WIPV4Address(1), WIPV4Address) is True)

	def test_bin_address(self):
		assert(len(WIPV4Address(0).bin_address()) == (4 * 8))
		assert(str(WIPV4Address("10.192.0.1").bin_address()) == "00001010110000000000000000000001")

	def test_from_string(self):
		pytest.raises(ValueError, WIPV4Address.from_string, '11.1.1')
		assert(str(WIPV4Address.from_string("10.192.0.1").bin_address()) == "00001010110000000000000000000001")

	def test_to_string(self):
		pytest.raises(TypeError, WIPV4Address.to_string, '10.192.0.1')
		address = WIPV4Address(WBinArray(0b00001010110000000000000000000001))
		assert(WIPV4Address.to_string(address) == "10.192.0.1")
		assert(WIPV4Address.to_string(address, dns_format=True) == "1.0.192.10.in-addr.arpa")

	def test_str(self):
		assert(str(WIPV4Address(WBinArray(0b00001010110000000000000000000001))) == "10.192.0.1")

	def test_bytes(self):
		assert(bytes(WIPV4Address(WBinArray(0b00001010110000000000000000000001))) == b"\x0a\xc0\x00\x01")


class TestNetworkIPV4:

	def test_init(self):
		assert(isinstance(WNetworkIPV4('10.0.0.0/24'), WNetworkIPV4) is True)
		pytest.raises(ValueError, WNetworkIPV4, '10.0.0.0/33')
		assert(isinstance(WNetworkIPV4('0.0.0.0/0'), WNetworkIPV4) is True)
		pytest.raises(ValueError, WNetworkIPV4, '10.0.0.0/-1')
		pytest.raises(ValueError, WNetworkIPV4, '10.0.0.1/24')
		assert(isinstance(WNetworkIPV4('10.0.0.1/24', network_address=False), WNetworkIPV4) is True)
		assert(isinstance(WNetworkIPV4((WIPV4Address('192.168.0.0'), 24)), WNetworkIPV4) is True)

	def test_address(self):
		assert(isinstance(WNetworkIPV4('10.0.0.0/24').address(), WIPV4Address) is True)
		assert(int(WNetworkIPV4('0.0.1.0/24').address().bin_address()) == 256)
		assert(int(WNetworkIPV4((WIPV4Address('0.0.1.0'), 24)).address().bin_address()) == 256)

	def test_mask(self):
		assert(isinstance(WNetworkIPV4('10.0.0.0/24').mask(), int) is True)
		assert(int(WNetworkIPV4('0.0.1.0/24').mask()) == 24)

	def test_first_address(self):
		assert(str(WNetworkIPV4('10.0.0.0/24').first_address()) == '10.0.0.1')
		assert(str(WNetworkIPV4('10.0.0.0/24').first_address(skip_network_address=False)) == '10.0.0.0')
		assert(str(WNetworkIPV4('10.0.0.1/24', network_address=False).first_address()) == '10.0.0.1')
		net = WNetworkIPV4('10.0.0.1/24', network_address=False)
		assert(str(net.first_address(skip_network_address=False)) == '10.0.0.0')

	def test_last_address(self):
		assert(str(WNetworkIPV4('10.0.0.0/23').last_address()) == '10.0.1.254')
		assert(str(WNetworkIPV4('10.0.0.0/23').last_address(skip_broadcast_address=False)) == '10.0.1.255')
		assert(str(WNetworkIPV4('10.0.0.1/23', network_address=False).last_address()) == '10.0.1.254')
		net = WNetworkIPV4('10.0.0.1/23', network_address=False)
		assert(str(net.last_address(skip_broadcast_address=False)) == '10.0.1.255')

	def test_iterator(self):
		assert(len(list(WNetworkIPV4('10.0.0.0/32').iterator())) == 1)
		assert(str(list(WNetworkIPV4('10.0.0.0/32').iterator())[0]) == '10.0.0.0')
		assert(len(list(WNetworkIPV4('10.0.0.0/32').iterator(skip_network_address=False))) == 1)
		assert(len(list(WNetworkIPV4('10.0.0.0/32').iterator(skip_broadcast_address=False))) == 1)
		net = WNetworkIPV4('10.0.0.0/32')
		assert(len(list(net.iterator(skip_broadcast_address=False, skip_network_address=False))) == 1)

		assert(len(list(WNetworkIPV4('10.0.0.0/31').iterator())) == 2)
		assert(str(list(WNetworkIPV4('10.0.0.0/31').iterator())[0]) == '10.0.0.0')
		assert(str(list(WNetworkIPV4('10.0.0.0/31').iterator())[1]) == '10.0.0.1')
		assert(len(list(WNetworkIPV4('10.0.0.0/31').iterator(skip_network_address=False))) == 2)
		assert(len(list(WNetworkIPV4('10.0.0.0/31').iterator(skip_broadcast_address=False))) == 2)
		net = WNetworkIPV4('10.0.0.0/31')
		assert(len(list(net.iterator(skip_broadcast_address=False, skip_network_address=False))) == 2)

		assert(len(list(WNetworkIPV4('10.0.0.0/28').iterator())) == 14)
		assert(str(list(WNetworkIPV4('10.0.0.0/28').iterator())[0]) == '10.0.0.1')
		assert(str(list(WNetworkIPV4('10.0.0.0/28').iterator(skip_network_address=False))[0]) == '10.0.0.0')
		assert(str(list(WNetworkIPV4('10.0.0.0/28').iterator())[13]) == '10.0.0.14')
		assert(str(list(WNetworkIPV4('10.0.0.0/28').iterator(skip_broadcast_address=False))[14]) == '10.0.0.15')
		assert(len(list(WNetworkIPV4('10.0.0.0/28').iterator(skip_network_address=False))) == 15)
		assert(len(list(WNetworkIPV4('10.0.0.0/28').iterator(skip_broadcast_address=False))) == 15)
		net = WNetworkIPV4('10.0.0.0/28')
		assert(len(list(net.iterator(skip_broadcast_address=False, skip_network_address=False))) == 16)

		it = WNetworkIPV4Iterator(WIPV4Address('10.0.0.0'))
		assert(len(list(it)) == 1)

	def test_contains(self):
		assert((WIPV4Address('192.168.0.1') in WNetworkIPV4('192.168.0.0/24')) is True)
		assert((WIPV4Address('192.168.0.1') in WNetworkIPV4('192.0.0.0/24')) is False)
		assert((WIPV4Address('192.168.0.0') in WNetworkIPV4('192.0.0.0/24')) is False)

	def test_multicast(self):
		assert(WNetworkIPV4.is_multicast(WIPV4Address('192.168.0.1')) is False)
		assert(WNetworkIPV4.is_multicast(WIPV4Address('239.200.200.1')) is True)


class TestWFQDN:
	def test_fqdn(self):
		assert(str(WFQDN() == ''))
		assert(str(WFQDN('site1.org2.fld.') == 'site1.org2.fld'))
		assert(str(WFQDN(['site1', 'org2', 'fld', '']) == 'site1.org2.fld'))

		pytest.raises(ValueError, WFQDN, 'a' * 1000)
		pytest.raises(ValueError, WFQDN, '\ud13f')

		assert(WFQDN.to_string(WFQDN('site1.org2.fld'), leading_dot=True) == 'site1.org2.fld.')
		pytest.raises(TypeError, WFQDN.to_string, 'site1.org2.fld')
		assert(WFQDN.to_string(WFQDN.from_string(''), leading_dot=True) == '.')

		assert(isinstance(WFQDN.punycode('\ud13f'), WFQDN) is True)
		assert(str(WFQDN.punycode('\ud13f')) == 'xn--3v7b')


class TestWIPPort:

	def test_port(self):
		pytest.raises(ValueError, WIPPort, 0)
		pytest.raises(ValueError, WIPPort, 70000)

		assert(int(WIPPort(70)) == 70)
		assert(str(WIPPort(70)) == "70")


class TestWIPV4SocketInfo:

	def test_socket_info(self):
		si = WIPV4SocketInfo('192.168.0.200', 2020)
		assert(isinstance(si.address(), WIPV4Address) is True)
		assert(str(si.address()) == '192.168.0.200')
		assert(isinstance(si.port(), WIPPort) is True)
		assert(int(si.port()) == 2020)

		assert(si.pair() == ('192.168.0.200', 2020))
		assert(WIPV4SocketInfo('10.20.30.40').pair() == ('10.20.30.40', 0))
		assert(WIPV4SocketInfo(None, 80).pair() == ('', 80))
		assert(WIPV4SocketInfo().pair() == ('', 0))

		assert(WIPV4SocketInfo(1).pair() == ('0.0.0.1', 0))

		assert(WIPV4SocketInfo(WFQDN('site1.org')).pair() == ('site1.org', 0))
		assert(WIPV4SocketInfo('site1.org').pair() == ('site1.org', 0))

		pytest.raises(ValueError, WIPV4SocketInfo, '!')

		assert(WIPV4SocketInfo('').address() is None)
		assert(WIPV4SocketInfo('').pair() == ('', 0))

		result = WIPV4SocketInfo.parse_socket_info('127.1.2.3:1020')
		assert(isinstance(result, WIPV4SocketInfo) is True)
		assert(str(result.address()) == '127.1.2.3')
		assert(int(result.port()) == 1020)

		result = WIPV4SocketInfo.parse_socket_info('127.1.2.5')
		assert(isinstance(result, WIPV4SocketInfo) is True)
		assert(str(result.address()) == '127.1.2.5')
		assert(result.port() is None)

		pytest.raises(ValueError, WIPV4SocketInfo.parse_socket_info, '127.1.2.5:11:')



# -*- coding: utf-8 -*-

import pytest
import io

from wasp_general.crypto.aes import WAESMode, WAES, WZeroPadding
from wasp_general.io import WAESWriter


class TestWAESWriter:

	def test(self):
		bytes_io = io.BytesIO()
		bytes_io.close = lambda: None  # do not really close the buffer

		secret_key = b'\x01\x01\x01\x01\x02\x02\x02\x02\x03\x03\x03\x03\x04\x04\x04\x04'
		iv = b'\x05\x05\x05\x05\x06\x06\x06\x06\x07\x07\x07\x07\x08\x08\x08\x08'
		aes_mode = WAESMode(16, 'AES-CBC', secret_key + iv)
		pytest.raises(ValueError, WAESWriter, bytes_io, WAES(aes_mode))

		aes_mode = WAESMode(16, 'AES-CBC', secret_key + iv, padding=WZeroPadding())
		aes = WAES(aes_mode)

		wr = WAESWriter(bytes_io, aes)
		assert(wr.write(b'bla-bla-bla long-long-long data very long data \x11\x11\x11\x11!!!') == 54)
		wr.close()
		bytes_io.seek(0)
		result = b'[C\xf0-\x9c\x98\x07-q\xc2?\xd8\xdeXLsn\x9cz\x00;\xa6b\x9b&_,\xfd\x91\xbe'
		result += b'\x03{\x91r(\xb0\xe7:\x81X\xf4\xe8(\x84\x9aM\x10\xc5\xf4\x01\xa9G\'\x04;`\r2\x17\x11'
		result += b'\x1b\xaf\xb4\x83'
		assert(bytes_io.read() == result)
		io.BytesIO.close(bytes_io)

		bytes_io = io.BytesIO()
		bytes_io.close = lambda: None  # do not really close the buffer
		wr = WAESWriter(bytes_io, aes)
		assert(wr.write(b'bla-bla-bla') == 11)
		wr.close()

		bytes_io.seek(0)

		result = b'\'\x16\x02\xd8\xd9\x19\x06\xaf/\x18]r\xed\x8f8\xe2'
		assert(bytes_io.read() == result)
		assert(aes.decrypt(result) == b'bla-bla-bla')

		aes_mode = WAESMode(16, 'AES-CBC', secret_key + iv)
		aes = WAES(aes_mode)
		assert(aes.decrypt(result) == b'bla-bla-bla\x00\x00\x00\x00\x00')

		io.BytesIO.close(bytes_io)



# -*- coding: utf-8 -*-
# wasp_general/composer.py
#
# Copyright (C) 2017 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code
# TODO: write tests for the code

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from abc import ABCMeta, abstractmethod

from wasp_general.verify import verify_type, verify_value, verify_subclass


class WComposerProto(metaclass=ABCMeta):

	@abstractmethod
	def compose(self, obj_spec):
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	def decompose(self, obj):
		raise NotImplementedError('This method is abstract')


class WPlainComposer(WComposerProto):

	@verify_subclass(strict_cls=(int, float, str, None))
	def __init__(self, strict_cls=None, permit_none=False):
		WComposerProto.__init__(self)
		self.__strict_cls = strict_cls
		self.__permit_none = permit_none

	def strict_cls(self):
		return self.__strict_cls

	def permit_none(self):
		return self.__permit_none

	@verify_type(obj_spec=(int, float, str, None))
	def compose(self, obj_spec):
		self.__check(obj_spec)
		return obj_spec

	@verify_type(obj_spec=(int, float, str, None))
	def decompose(self, obj):
		self.__check(obj)
		return obj

	def __check(self, obj):
		if obj is None:
			if self.permit_none() is False:
				raise TypeError('None value spotted')
		else:
			strict_cls = self.strict_cls()
			if strict_cls is not None and isinstance(obj, strict_cls) is False:
				raise TypeError(
					'Invalid type ("%s" should be "%s")' %
					(obj.__class__.__name__, strict_cls.__name__)
				)


# noinspection PyAbstractClass
class WProxyComposer(WComposerProto):

	@verify_type(basic_composer=WComposerProto)
	def __init__(self, basic_composer):
		WComposerProto.__init__(self)
		self.__basic_composer = basic_composer

	def basic_composer(self):
		return self.__basic_composer


class WIterComposer(WProxyComposer):

	@verify_type(obj_spec=(list, set, tuple))
	def compose(self, obj_spec):
		return [self.basic_composer().compose(x) for x in obj_spec]

	@verify_type(obj_spec=(list, set, tuple))
	def decompose(self, obj):
		return [self.basic_composer().decompose(x) for x in obj]


class WCompositeComposer(WComposerProto):

	# noinspection PyAbstractClass
	class CompositeKey(WProxyComposer):

		@verify_type('paranoid', basic_composer=WComposerProto)
		@verify_type(required=bool)
		def __init__(self, key, basic_composer, required=False):
			WProxyComposer.__init__(self, basic_composer)
			self.__key = key
			self.__required = required

		def key(self):
			return self.__key

		def required(self):
			return self.__required

		def compose(self, key_value):
			return self.basic_composer().compose(key_value)

		def decompose(self, key_value):
			return self.basic_composer().decompose(key_value)

		@abstractmethod
		def has_key(self, obj):
			raise NotImplementedError('This method is abstract')

		@abstractmethod
		def get_key(self, obj):
			raise NotImplementedError('This method is abstract')

		@abstractmethod
		def set_key(self, obj, value):
			raise NotImplementedError('This method is abstract')

	class KeyInstance:

		def __init__(self, composite_key, value):
			if isinstance(composite_key, WCompositeComposer.CompositeKey) is False:
				raise TypeError('Invalid composite_key type')

			self.composite_key = composite_key
			self.value = value

	class ConstructionKeys:

		def __init__(self, *key_instances):
			self.__key_instances = []
			for instance in key_instances:
				self.add(instance)

		def add(self, key_instance):
			if isinstance(key_instance, WCompositeComposer.KeyInstance) is False:
				raise TypeError('Invalid construction pair type')
			self.__key_instances.append(key_instance)

		def key_instances(self):
			return self.__key_instances.copy()

		def __getitem__(self, key):
			for instance in self.key_instances():
				if instance.composite_key.key() == key:
					return instance
			raise KeyError('Invalid key: "%s"' % str(key))

		def has_key(self, key):
			for instance in self.key_instances():
				if instance.composite_key.key() == key:
					return True
			return False

		def __iter__(self):
			for pair in self.key_instances():
				yield pair

	class InstanceConstructor(metaclass=ABCMeta):

		def construct_obj(self, construction_keys):
			if isinstance(construction_keys, WCompositeComposer.ConstructionKeys) is False:
				raise TypeError('Invalid construction_keys type')

			result = self.create_obj(construction_keys)
			for instance in construction_keys:
				instance.composite_key.set_key(result, instance.value)
			return result

		@abstractmethod
		def create_obj(self, construction_keys):
			raise NotImplementedError('This method is abstract')

	@verify_type('paranoid', composite_keys=CompositeKey)
	@verify_type(constructor=(InstanceConstructor, None))
	def __init__(self, *composite_keys, constructor=None):
		self.__constructor = \
			constructor if constructor is not None else WCompositeComposer.InstanceConstructor()
		self.__required_keys = []
		self.__optional_keys = []

		for key in composite_keys:
			self.add_composite_key(key)

	def constructor(self):
		return self.__constructor

	@verify_type(composite_key=CompositeKey)
	def add_composite_key(self, composite_key):
		keys_list = self.__required_keys if composite_key.required() else self.__optional_keys
		keys_list.append(composite_key)

	def required_keys(self):
		return self.__required_keys.copy()

	def optional_keys(self):
		return self.__optional_keys.copy()

	@verify_type(obj_spec=dict)
	def compose(self, obj_spec):
		obj_spec = obj_spec.copy()
		construction_keys = WCompositeComposer.ConstructionKeys()

		for composite_key in self.required_keys():
			key = composite_key.key()
			if key not in obj_spec.keys():
				raise TypeError('Required key not found')
			value = composite_key.compose(obj_spec[key])
			obj_spec.pop(key)
			construction_keys.add(WCompositeComposer.KeyInstance(composite_key, value))

		for composite_key in self.optional_keys():
			key = composite_key.key()
			if key in obj_spec.keys():
				value = composite_key.compose(obj_spec[key])
				obj_spec.pop(key)
				construction_keys.add(WCompositeComposer.KeyInstance(composite_key, value))

		if len(obj_spec) > 0:
			raise TypeError('Unable to process every fields')

		return self.constructor().construct_obj(construction_keys)

	def decompose(self, obj):
		result = {}

		for composite_key in self.required_keys():
			if composite_key.has_key(obj) is False:
				raise TypeError('Required key not found')
			value = composite_key.decompose(composite_key.get_key(obj))
			result[composite_key.key()] = value

		for composite_key in self.optional_keys():
			if composite_key.has_key(obj) is True:
				value = composite_key.decompose(composite_key.get_key(obj))
				result[composite_key.key()] = value

		return result


class WDictComposer(WCompositeComposer):

	class DictKey(WCompositeComposer.CompositeKey):

		def has_key(self, obj):
			return self.key() in obj.keys()

		def get_key(self, obj):
			return obj[self.key()]

		def set_key(self, obj, value):
			obj[self.key()] = value
			return obj

	class DictConstructor(WCompositeComposer.InstanceConstructor):

		def create_obj(self, construction_keys):
			return {}

	@verify_type('paranoid', composite_keys=DictKey)
	def __init__(self, *composite_keys):
		WCompositeComposer.__init__(self, *composite_keys, constructor=WDictComposer.DictConstructor())

	@verify_type(composite_key=DictKey)
	def add_composite_key(self, composite_key):
		WCompositeComposer.add_composite_key(self, composite_key)


class WClassComposer(WCompositeComposer):

	class ClassKey(WCompositeComposer.CompositeKey):

		@verify_type('paranoid', basic_composer=WComposerProto, required=bool)
		@verify_type(key=str)
		@verify_value(key=lambda x: len(x) > 0)
		@verify_value(has_key_fn=lambda x: x is None or callable(x))
		@verify_value(get_key_fn=lambda x: x is None or callable(x))
		@verify_value(set_key_fn=lambda x: x is None or callable(x))
		def __init__(
			self, key, basic_composer, has_key_fn=None, get_key_fn=None, set_key_fn=None, required=False
		):
			WCompositeComposer.CompositeKey.__init__(self, key, basic_composer, required=required)
			self.__get_key_fn = get_key_fn
			self.__set_key_fn = set_key_fn
			self.__has_key_fn = has_key_fn

		def has_key(self, obj):
			if self.__has_key_fn is not None:
				return self.__has_key_fn(obj, self.key())
			return hasattr(obj, self.key())

		def get_key(self, obj):
			if self.__get_key_fn is not None:
				return self.__get_key_fn(obj, self.key())
			return getattr(obj, self.key())

		def set_key(self, obj, value):
			if self.__set_key_fn is not None:
				return self.__set_key_fn(obj, self.key(), value)
			setattr(obj, self.key(), value)
			return obj

	class GetterKey(ClassKey):

		def __init__(self, key, basic_composer, required=False):
			WClassComposer.ClassKey.__init__(self, key, basic_composer=basic_composer, required=required)

		def get_key(self, obj):
			return WClassComposer.ClassKey.get_key(self, obj)()

		def set_key(self, obj, value):
			return obj

	class ClassConstructor(WCompositeComposer.InstanceConstructor):

		@verify_type(basic_cls=type)
		@verify_value(create_obj_fn=(lambda x: x is None or callable(x)))
		def __init__(self, basic_cls, create_obj_fn=None):
			self.__basic_cls = basic_cls
			self.__create_obj_fn = create_obj_fn

		def basic_cls(self):
			return self.__basic_cls

		def create_obj(self, construction_keys):
			if self.__create_obj_fn is not None:
				return self.__create_obj_fn(construction_keys)
			return self.basic_cls()()

	@verify_type('paranoid', composite_keys=ClassKey)
	@verify_type(constructor=ClassConstructor)
	def __init__(self, *composite_keys, constructor=None):
		WCompositeComposer.__init__(self, *composite_keys, constructor=constructor)

	@verify_type(composite_key=ClassKey)
	def add_composite_key(self, composite_key):
		WCompositeComposer.add_composite_key(self, composite_key)

	def basic_cls(self):
		return self.constructor().basic_cls()


class WComposerFactory(WComposerProto):

	class Entry:

		@verify_type(composer=WClassComposer, name=(str, None))
		@verify_value(name=lambda x: x is None or len(x) > 0)
		def __init__(self, composer, name=None):
			self.__composer = composer
			self.__name = name if name is not None else composer.basic_cls().__name__

		def composer(self):
			return self.__composer

		def name(self):
			return self.__name

	__default_name_field__ = '__cls__'
	__default_value_field__ = '__instance__'

	@verify_type('paranoid', entries=Entry)
	@verify_type(name_field=(str, None), value_field=(str, None))
	@verify_value(name_field=lambda x: x is None or len(x) > 0)
	@verify_value(value_field=lambda x: x is None or len(x) > 0)
	def __init__(self, *entries, name_field=None, value_field=None):
		WComposerProto.__init__(self)

		self.__entries = {}
		for entry in entries:
			self.add_entry(entry)

		self.__name_field = name_field if name_field is not None else self.__default_name_field__
		self.__value_field = value_field if value_field is not None else self.__default_value_field__

	@verify_type(entry=Entry)
	def add_entry(self, entry):
		entry_name = entry.name()
		if entry_name in self.__entries.keys():
			raise RuntimeError('Multiple entries with the same name "%s" spotted' % entry_name)
		self.__entries[entry_name] = entry

	def entries(self):
		return self.__entries.copy()

	def name_field(self):
		return self.__name_field

	def value_field(self):
		return self.__value_field

	@verify_type(obj_spec=dict)
	def compose(self, obj_spec):
		name_field = self.name_field()
		value_field = self.value_field()
		if len(obj_spec) != 2 or name_field not in obj_spec.keys() or value_field not in obj_spec.keys():
			raise TypeError('Data malformed')

		entries = self.entries()
		entry_name = obj_spec[name_field]
		entry_value = obj_spec[value_field]

		if entry_name not in entries.keys():
			raise TypeError('Invalid class "%s" was specified' % entry_name)

		entry = entries[entry_name]
		return entry.composer().compose(entry_value)

	@verify_type(obj=object)
	def decompose(self, obj):

		suitable_entries = []
		for entry in self.entries().values():
			if isinstance(obj, entry.composer().basic_cls()) is True:
				suitable_entries.append(entry)

		entries_count = len(suitable_entries)
		if entries_count == 0:
			raise TypeError(
				'Unable to find suitable entry for "%s" (no one matched)' % obj.__class__.__name__
			)
		elif entries_count > 1:
			reduced_subcls_entries = []
			for i_entry in suitable_entries:
				is_subcls = False
				i_subcls = i_entry.composer().basic_cls()
				for j_entry in suitable_entries:
					j_subcls = j_entry.composer().basic_cls()
					if issubclass(i_subcls, j_subcls) is True and i_subcls != j_subcls:
						is_subcls = True
						break
				if is_subcls is False:
					reduced_subcls_entries.append(i_entry)

			subcls_entries_count = len(reduced_subcls_entries)
			if subcls_entries_count == 0:
				raise TypeError('Unable to find suitable entry for "%s"' % obj.__class__.__name__)
			elif subcls_entries_count > 1:
				raise TypeError(
					'Unable to find suitable entry for "%s" (matched two or more entries)' %
					obj.__class__.__name__
				)
			suitable_entries = reduced_subcls_entries

		entry = suitable_entries[0]
		decompose_result = {
			self.name_field(): entry.name(),
			self.value_field(): entry.composer().decompose(obj)
		}
		return decompose_result



# -*- coding: utf-8 -*-

import pytest

from wasp_general.os.linux.mounts import WMountPoint


class TestWMountPoint:

	class NullFileMP(WMountPoint):
		__mounts_file__ = '/dev/null'

	__mounts_file_example__ = """/dev/sda1 / ext4 rw,noatime,discard,data=ordered 0 0
devtmpfs /dev devtmpfs rw,relatime,size=5854588k,nr_inodes=1463647,mode=755 0 0
sysfs /sys sysfs rw,nosuid,nodev,noexec,relatime 0 0
proc /proc proc rw,nosuid,nodev,noexec,relatime 0 0
tmpfs /dev/shm tmpfs rw,nosuid,nodev 0 0
devpts /dev/pts devpts rw,nosuid,noexec,relatime,gid=5,mode=620,ptmxmode=000 0 0
tmpfs /run tmpfs rw,nosuid,nodev,mode=755 0 0
"""

	def test(self, temp_file):
		mp = WMountPoint('/dev/sda1 / ext4 rw,noatime,discard,data=ordered 1 1')
		assert(isinstance(mp, WMountPoint) is True)
		assert(mp.device() == '/dev/sda1')
		assert(mp.device_name() == 'sda1')
		assert(mp.path() == '/')
		assert(mp.fs() == 'ext4')
		assert(mp.options() == ('rw', 'noatime', 'discard', 'data=ordered'))
		assert(mp.dump_flag() is True)
		assert(mp.fsck_order() == 1)

		pytest.raises(RuntimeError, WMountPoint, '')
		pytest.raises(RuntimeError, WMountPoint, '/dev/ fat32')
		pytest.raises(RuntimeError, WMountPoint, '/dev/sda1 / ext4 rw,noatime,discard,data=ordered 2 1')

		assert(TestWMountPoint.NullFileMP.mounts() == tuple())

		with open(temp_file, 'w') as f:
			f.write(TestWMountPoint.__mounts_file_example__)

		class MountsExample(WMountPoint):
			__mounts_file__ = temp_file

		mounts = MountsExample.mounts()
		assert(len(mounts) == 7)

		assert(isinstance(mounts[0], WMountPoint) is True)
		assert(mounts[0].device() == '/dev/sda1')

		assert(isinstance(mounts[1], WMountPoint) is True)
		assert(mounts[1].device_name() == 'devtmpfs')

		assert(isinstance(mounts[2], WMountPoint) is True)
		assert(mounts[2].path() == '/sys')

		assert(isinstance(mounts[3], WMountPoint) is True)
		assert(mounts[3].fs() == 'proc')

		assert(isinstance(mounts[4], WMountPoint) is True)
		assert(mounts[4].options() == ('rw', 'nosuid', 'nodev'))

		assert(isinstance(mounts[5], WMountPoint) is True)
		assert(mounts[5].dump_flag() is False)

		assert(isinstance(mounts[6], WMountPoint) is True)
		assert(mounts[6].fsck_order() == 0)

		assert(TestWMountPoint.NullFileMP.mount_point('/foo') is None)

		mp = MountsExample.mount_point('/foo/bar')
		assert(isinstance(mp, WMountPoint) is True)
		assert(mp.device() == '/dev/sda1')

		mp = MountsExample.mount_point('/dev/file1')
		assert(isinstance(mp, WMountPoint) is True)
		assert(mp.device() == 'devtmpfs')

		mp = MountsExample.mount_point('/dev/pts/file1')
		assert(isinstance(mp, WMountPoint) is True)
		assert(mp.device() == 'devpts')



# -*- coding: utf-8 -*-
# wasp_general/crypto/hex.py
#
# Copyright (C) 2016 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from binascii import hexlify, unhexlify

from wasp_general.verify import verify_type


class WHex:
	""" binascii.hexlify wrapper. Converts bytes to hex-string
	"""

	@verify_type(byte_sequence=bytes)
	def __init__(self, byte_sequence):
		""" Create converter

		:param byte_sequence: sequence to convert
		"""

		self.__byte_sequence = byte_sequence

	def __str__(self):
		""" Return result of converting the sequence

		:return: str
		"""
		return hexlify(self.__byte_sequence).decode('ascii')


class WUnHex:
	""" binascii.unhexlify wrapper. Converts string to bytes
	"""

	@verify_type(string=(str, bytes))
	def __init__(self, string):
		""" Create converter

		:param string: hex-string to convert
		"""

		self.__string = string

	def __bytes__(self):
		""" Return result of converting the hex-string

		:return: bytes
		"""
		return unhexlify(self.__string)
